#!/usr/bin/env python3
"""
Search tool for Local Search MCP Server (SSOT).
"""
import os
import time
from typing import Any, Dict, List

from sari.mcp.tools._util import (
    mcp_response,
    pack_header,
    pack_line,
    pack_truncated,
    pack_encode_id,
    pack_encode_text,
    resolve_root_ids,
    pack_error,
    ErrorCode,
)

from sari.core.db import LocalSearchDB, SearchOptions
from sari.core.engine_runtime import EngineError, SqliteSearchEngineAdapter
from sari.mcp.telemetry import TelemetryLogger


def execute_search(
    args: Dict[str, Any],
    db: LocalSearchDB,
    logger: TelemetryLogger,
    roots: List[str],
    engine: Any = None,
) -> Dict[str, Any]:
    start_ts = time.time()
    engine = engine or getattr(db, "engine", None)

    root_ids = resolve_root_ids(roots)
    req_root_ids = args.get("root_ids")
    if isinstance(req_root_ids, list):
        req_root_ids = [str(r) for r in req_root_ids if r]
        if root_ids:
            root_ids = [r for r in root_ids if r in req_root_ids]
        else:
            root_ids = list(req_root_ids)
        if req_root_ids and not root_ids:
            if db and db.has_legacy_paths():
                root_ids = []
            else:
                return mcp_response(
                    "search",
                    lambda: pack_error("search", ErrorCode.ERR_ROOT_OUT_OF_SCOPE, "root_ids out of scope", hints=["outside final_roots"]),
                    lambda: {"error": {"code": ErrorCode.ERR_ROOT_OUT_OF_SCOPE.value, "message": "root_ids out of scope"}, "isError": True},
                )

    query = (args.get("query") or "").strip()
    if not query:
        return mcp_response(
            "search",
            lambda: pack_error("search", ErrorCode.INVALID_ARGS, "query is required"),
            lambda: {"error": {"code": ErrorCode.INVALID_ARGS.value, "message": "query is required"}, "isError": True},
        )

    repo = args.get("scope") or args.get("repo")
    if repo == "workspace":
        repo = None

    file_types = list(args.get("file_types", []))
    search_type = args.get("type")
    if search_type == "docs":
        doc_exts = ["md", "txt", "pdf", "docx", "rst", "pdf"]
        file_types.extend([e for e in doc_exts if e not in file_types])

    try:
        limit = int(args.get("limit", 8))
    except (ValueError, TypeError):
        limit = 8
    limit = max(1, min(limit, 50))

    try:
        offset = max(int(args.get("offset", 0)), 0)
    except (ValueError, TypeError):
        offset = 0

    try:
        raw_lines = int(args.get("context_lines", 5))
        snippet_lines = min(max(raw_lines, 1), 20)
    except (ValueError, TypeError):
        snippet_lines = 5

    total_mode = str(args.get("total_mode") or "").strip().lower()
    if total_mode not in {"exact", "approx"}:
        total_mode = "exact"

    engine_mode = "sqlite"
    index_version = ""
    if engine and hasattr(engine, "status"):
        st = engine.status()
        engine_mode = st.engine_mode
        index_version = st.index_version
        if engine_mode == "embedded" and not st.engine_ready:
            if st.reason == "NOT_INSTALLED":
                auto_install = (os.environ.get("SARI_ENGINE_AUTO_INSTALL", "1").strip().lower() not in {"0", "false", "no", "off"})
                if not auto_install:
                    engine = SqliteSearchEngineAdapter(db)
                    st = engine.status()
                    engine_mode = st.engine_mode
                    index_version = st.index_version
                if hasattr(engine, "install"):
                    try:
                        engine.install()
                        st = engine.status()
                        engine_mode = st.engine_mode
                        index_version = st.index_version
                    except EngineError as exc:
                        engine = SqliteSearchEngineAdapter(db)
                        st = engine.status()
                        engine_mode = st.engine_mode
                        index_version = st.index_version
                    except Exception as exc:
                        engine = SqliteSearchEngineAdapter(db)
                        st = engine.status()
                        engine_mode = st.engine_mode
                        index_version = st.index_version
            if engine_mode == "embedded" and not st.engine_ready:
                engine = SqliteSearchEngineAdapter(db)
                st = engine.status()
                engine_mode = st.engine_mode
                index_version = st.index_version
    elif engine is None:
        engine = SqliteSearchEngineAdapter(db)
        st = engine.status()
        engine_mode = st.engine_mode
        index_version = st.index_version

    opts = SearchOptions(
        query=query,
        repo=repo,
        limit=limit,
        offset=offset,
        snippet_lines=snippet_lines,
        file_types=file_types,
        path_pattern=args.get("path_pattern"),
        exclude_patterns=args.get("exclude_patterns", []),
        recency_boost=bool(args.get("recency_boost", False)),
        use_regex=bool(args.get("use_regex", False)),
        case_sensitive=bool(args.get("case_sensitive", False)),
        total_mode=total_mode,
        root_ids=root_ids,
    )

    try:
        hits, meta = engine.search_v2(opts) if engine else ([], {})
    except EngineError as exc:
        code = getattr(ErrorCode, exc.code, ErrorCode.ERR_ENGINE_QUERY)
        return mcp_response(
            "search",
            lambda: pack_error("search", code, exc.message, hints=[exc.hint] if exc.hint else None),
            lambda: {"error": {"code": code.value, "message": exc.message, "hint": exc.hint}, "isError": True},
        )
    except Exception as exc:
        return mcp_response(
            "search",
            lambda: pack_error("search", ErrorCode.ERR_ENGINE_QUERY, f"engine query failed: {exc}"),
            lambda: {"error": {"code": ErrorCode.ERR_ENGINE_QUERY.value, "message": f"engine query failed: {exc}"}, "isError": True},
        )

    latency_ms = int((time.time() - start_ts) * 1000)
    total = meta.get("total", -1)
    total_mode = meta.get("total_mode", total_mode)

    def build_json() -> Dict[str, Any]:
        results: List[Dict[str, Any]] = []
        for hit in hits:
            results.append({
                "doc_id": hit.path,
                "repo": hit.repo,
                "path": hit.path,
                "score": hit.score,
                "snippet": hit.snippet,
                "mtime": hit.mtime,
                "size": hit.size,
                "match_count": hit.match_count,
                "file_type": hit.file_type,
                "hit_reason": hit.hit_reason,
                "context_symbol": hit.context_symbol,
                "docstring": hit.docstring,
                "metadata": hit.metadata,
            })
        return {
            "query": query,
            "limit": limit,
            "offset": offset,
            "results": results,
            "meta": {
                "total": total,
                "total_mode": total_mode,
                "engine": engine_mode,
                "latency_ms": latency_ms,
                "index_version": index_version,
            },
        }

    def build_pack() -> str:
        returned = len(hits)
        header = pack_header("search", {"q": pack_encode_text(query)}, returned=returned)
        lines = [header]
        lines.append(pack_line("m", {"total": str(total)}))
        lines.append(pack_line("m", {"total_mode": total_mode}))
        lines.append(pack_line("m", {"engine": engine_mode}))
        lines.append(pack_line("m", {"latency_ms": str(latency_ms)}))
        if index_version:
            lines.append(pack_line("m", {"index_version": pack_encode_id(index_version)}))
        for h in hits:
            lines.append(pack_line("r", {
                "path": pack_encode_id(h.path),
                "repo": pack_encode_id(h.repo),
                "score": f"{h.score:.3f}",
                "mtime": str(h.mtime),
                "size": str(h.size),
                "file_type": pack_encode_id(h.file_type),
                "snippet": pack_encode_text(h.snippet),
            }))
        if returned >= limit:
            lines.append(pack_truncated(offset + limit, limit, "maybe"))
        return "\n".join(lines)

    return mcp_response("search", build_pack, build_json)



import pytest
import os
import glob
import importlib
import inspect
from unittest.mock import MagicMock

def test_bulk_execute_tools():
    # Only test a subset if it takes too long, or optimize
    tool_files = glob.glob("sari/mcp/tools/*.py")
    db = MagicMock()
    indexer = MagicMock()
    logger = MagicMock()
    roots = ["/tmp/ws"]
    
    db.get_repo_stats.return_value = {"repo1": 1}
    db.list_files.return_value = ([], {"total": 0})
    db.search_files.return_value = []
    db.read_file.return_value = "content"
    
    for f in tool_files:
        if f.endswith("__init__.py") or f.endswith("_util.py"):
            continue
            
        module_name = f.replace("/", ".").replace(".py", "")
        if module_name.startswith("sari.sari."):
            module_name = module_name[5:]
            
        try:
            mod = importlib.import_module(module_name)
            for name, func in inspect.getmembers(mod, inspect.isfunction):
                if name.startswith("execute_"):
                    sig = inspect.signature(func)
                    call_args = []
                    for p in sig.parameters.values():
                        p_name = p.name.lower()
                        if p_name == "args": call_args.append({"query": "t", "tag": "t", "path": "root-1/f"})
                        elif p_name == "db": call_args.append(db)
                        elif p_name == "indexer": call_args.append(indexer)
                        elif p_name == "logger": call_args.append(logger)
                        elif p_name == "roots": call_args.append(roots)
                        elif p_name == "cfg": call_args.append(MagicMock())
                        elif p_name == "workspace_root": call_args.append("/tmp/ws")
                        elif p_name == "server_version": call_args.append("0.0.1")
                        elif p_name == "engine": call_args.append(MagicMock())
                        else: call_args.append(None)
                    try:
                        # Skip if it's too slow or blocks
                        func(*call_args)
                    except Exception: pass
        except Exception: pass


import sqlite3
from typing import Any, Optional

class BaseRepository:
    def __init__(self, conn: sqlite3.Connection):
        self._conn = conn

    @property
    def connection(self) -> sqlite3.Connection:
        return self._conn

    def execute(self, sql: str, params: Any = None) -> sqlite3.Cursor:
        if params:
            return self._conn.execute(sql, params)
        return self._conn.execute(sql)



import json
import os
import time
import socket
import pytest
import subprocess
from pathlib import Path

class TestMultiWorkspaceIntegration:
    
    @pytest.fixture
    def multi_env(self, tmp_path):
        fake_home = tmp_path / "multi_home"
        fake_home.mkdir()
        
        env = os.environ.copy()
        env["HOME"] = str(fake_home)
        env["SARI_REGISTRY_FILE"] = str(tmp_path / "multi_registry.json")
        env["PYTHONPATH"] = os.getcwd() + ":" + env.get("PYTHONPATH", "")
        env["SARI_DAEMON_AUTOSTART"] = "1"
        env["SARI_DAEMON_PORT"] = "48100"
        
        return env

    def test_daemon_handles_multiple_workspaces(self, tmp_path, multi_env):
        ws_paths = []
        for i in range(3):
            ws = tmp_path / f"ws_{i}"
            ws.mkdir()
            (ws / ".sari").mkdir()
            (ws / f"file_{i}.txt").write_text(f"Content from WS {i}")
            ws_paths.append(str(ws.expanduser().resolve()))

        multi_env["SARI_WORKSPACE_ROOT"] = ws_paths[0]
        subprocess.run(["python3", "-m", "sari.mcp.cli", "daemon", "start", "-d", "--daemon-port", "48100"], env=multi_env, check=True)
        time.sleep(2.0)
        
        try:
            def send_init(ws_path):
                init_req = json.dumps({"jsonrpc":"2.0","id":1, "method":"initialize","params":{"rootUri": f"file://{ws_path}"}})
                frame = f"Content-Length: {len(init_req)}\r\n\r\n{init_req}".encode()
                with socket.create_connection(("127.0.0.1", 48100)) as sock:
                    sock.sendall(frame)
                    resp = sock.recv(4096)
                    assert b'"result":' in resp

            send_init(ws_paths[1])
            send_init(ws_paths[2])
            
            time.sleep(3.0) 
            
            from sari.core.server_registry import ServerRegistry
            os.environ["SARI_REGISTRY_FILE"] = multi_env["SARI_REGISTRY_FILE"]
            registry = ServerRegistry()
            
            ports = []
            for ws in ws_paths:
                info = registry.get_workspace(ws)
                assert info is not None, f"WS {ws} not registered"
                assert info.get("http_port") is not None
                ports.append(info["http_port"])
            
            assert len(set(ports)) == 3, f"Ports should be unique: {ports}"
            
            for i, port in enumerate(ports):
                with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
                    s.settimeout(1.0)
                    assert s.connect_ex(("127.0.0.1", port)) == 0, f"HTTP server for WS {i} at {port} not reachable"

        finally:
            subprocess.run(["python3", "-m", "sari.mcp.cli", "daemon", "stop", "--daemon-port", "48100"], env=multi_env)


import json
import threading
import time
from datetime import datetime
from typing import Any, Dict, List

from sari.mcp.tools._util import (
    mcp_response,
    pack_header,
    pack_line,
    pack_encode_id,
    pack_encode_text,
    pack_error,
    ErrorCode,
)

from sari.core.queue_pipeline import DbTask


def _enqueue_or_write(db: Any, indexer: Any, row: tuple) -> None:
    writer = getattr(indexer, "_db_writer", None) if indexer else None
    if writer and DbTask:
        writer.enqueue(DbTask(kind="upsert_contexts", context_rows=[row]))
        return
    prev = getattr(db, "_writer_thread_id", None)
    db.register_writer_thread(threading.get_ident())
    try:
        with db._lock:
            cur = db._write.cursor()
            cur.execute("BEGIN")
            db.upsert_context_tx(cur, [row])
            db._write.commit()
    finally:
        db.register_writer_thread(prev)


def build_archive_context(args: Dict[str, Any], db: Any, indexer: Any = None) -> Dict[str, Any]:
    topic = str(args.get("topic") or "").strip()
    content = str(args.get("content") or "").strip()
    tags = args.get("tags") or []
    related = args.get("related_files") or []
    source = str(args.get("source") or "").strip()
    valid_from_raw = args.get("valid_from")
    valid_until_raw = args.get("valid_until")
    deprecated = int(bool(args.get("deprecated"))) if args.get("deprecated") is not None else 0
    if not topic or not content:
        raise ValueError("topic and content are required")
    if isinstance(tags, str):
        tags = [t.strip() for t in tags.split(",") if t.strip()]
    if isinstance(related, str):
        related = [r.strip() for r in related.split(",") if r.strip()]
    tags_json = json.dumps(tags, ensure_ascii=False)
    related_json = json.dumps(related, ensure_ascii=False)
    def _parse_ts(v):
        if v is None or v == "":
            return 0
        if isinstance(v, (int, float)):
            return int(v)
        s = str(v).strip()
        if s.isdigit():
            return int(s)
        try:
            return int(datetime.fromisoformat(s).timestamp())
        except Exception:
            return 0
    valid_from = _parse_ts(valid_from_raw)
    valid_until = _parse_ts(valid_until_raw)
    now = int(time.time())
    row = (topic, content, tags_json, related_json, source, valid_from, valid_until, deprecated, now, now)
    _enqueue_or_write(db, indexer, row)
    return {"topic": topic, "tags": tags, "related_files": related, "source": source, "valid_from": valid_from, "valid_until": valid_until, "deprecated": deprecated}


def execute_archive_context(args: Dict[str, Any], db: Any, roots: List[str], indexer: Any = None) -> Dict[str, Any]:
    try:
        payload = build_archive_context(args, db, indexer=indexer)
    except ValueError as e:
        return mcp_response(
            "archive_context",
            lambda: pack_error("archive_context", ErrorCode.INVALID_ARGS, str(e)),
            lambda: {"error": {"code": ErrorCode.INVALID_ARGS.value, "message": str(e)}, "isError": True},
        )

    def build_pack() -> str:
        lines = [pack_header("archive_context", {"topic": pack_encode_text(payload["topic"])}, returned=1)]
        kv = {
            "topic": pack_encode_id(payload["topic"]),
            "tags": pack_encode_text(",".join(payload.get("tags", []))),
        }
        lines.append(pack_line("r", kv))
        return "\n".join(lines)

    return mcp_response(
        "archive_context",
        build_pack,
        lambda: payload,
    )



import json
import os
import subprocess
import time
import socket
import pytest
import shutil
from pathlib import Path
from unittest.mock import patch, MagicMock

class TestDaemonCliIntegration:
    
    @pytest.fixture
    def test_env(self, tmp_path):
        fake_home = tmp_path / "fake_home"
        fake_home.mkdir()
        
        env = os.environ.copy()
        env["HOME"] = str(fake_home)
        env["USERPROFILE"] = str(fake_home)
        env["SARI_REGISTRY_FILE"] = str(tmp_path / "registry.json")
        env["PYTHONPATH"] = os.getcwd() + ":" + env.get("PYTHONPATH", "")
        env["SARI_DAEMON_AUTOSTART"] = "1"
        env["SARI_DAEMON_IDLE_SEC"] = "60" # Default long enough
        env["SARI_DAEMON_HEARTBEAT_SEC"] = "1"
        
        return env

    @pytest.fixture
    def workspace(self, tmp_path):
        ws = tmp_path / "test_ws"
        ws.mkdir()
        (ws / ".sari").mkdir(parents=True, exist_ok=True)
        return str(ws.expanduser().resolve())

    def test_daemon_lifecycle(self, workspace, test_env):
        test_env["SARI_WORKSPACE_ROOT"] = workspace
        port = "47891"
        
        try:
            subprocess.run(
                ["python3", "-m", "sari.mcp.cli", "daemon", "start", "-d", "--daemon-port", port],
                env=test_env, check=True
            )
            time.sleep(2.0)
            
            result = subprocess.run(
                ["python3", "-m", "sari.mcp.cli", "daemon", "status", "--daemon-port", port],
                env=test_env, capture_output=True, text=True
            )
            assert "Running" in result.stdout
            
            subprocess.run(
                ["python3", "-m", "sari.mcp.cli", "daemon", "stop", "--daemon-port", port],
                env=test_env, check=True
            )
            time.sleep(0.5)
            
            result = subprocess.run(
                ["python3", "-m", "sari.mcp.cli", "daemon", "status", "--daemon-port", port],
                env=test_env, capture_output=True, text=True
            )
            assert "Stopped" in result.stdout
            
        finally:
            subprocess.run(["python3", "-m", "sari.mcp.cli", "daemon", "stop", "--daemon-port", port], env=test_env)

    def test_server_auto_proxy_mode(self, workspace, test_env):
        test_env["SARI_WORKSPACE_ROOT"] = workspace
        port = "47894"
        test_env["SARI_DAEMON_PORT"] = port
        
        subprocess.run(["python3", "-m", "sari.mcp.cli", "daemon", "start", "-d", "--daemon-port", port], env=test_env, check=True)
        time.sleep(2.5)
        
        try:
            init_req = json.dumps({"jsonrpc":"2.0","id":1,"method":"initialize","params":{"protocolVersion":"2025-11-25"}})
            test_env["SARI_DEV_JSONL"] = "1"
            
            proc = subprocess.Popen(
                ["python3", "-m", "sari.mcp.server"],
                env=test_env, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE
            )
            
            stdout, stderr = proc.communicate(input=(init_req + "\n").encode(), timeout=10)
            stdout_str = stdout.decode(errors='ignore')
            
            assert '"result":' in stdout_str
            
        finally:
            subprocess.run(["python3", "-m", "sari.mcp.cli", "daemon", "stop", "--daemon-port", port], env=test_env)

    def test_daemon_idle_auto_stop(self, workspace, test_env):
        test_env["SARI_WORKSPACE_ROOT"] = workspace
        port = "47896"
        test_env["SARI_DAEMON_IDLE_SEC"] = "2"
        test_env["SARI_DAEMON_IDLE_WITH_ACTIVE"] = "1"
        
        subprocess.run(["python3", "-m", "sari.mcp.cli", "daemon", "start", "-d", "--daemon-port", port], env=test_env, check=True)
        time.sleep(2.0)
        
        try:
            result = subprocess.run(["python3", "-m", "sari.mcp.cli", "daemon", "status", "--daemon-port", port], env=test_env, capture_output=True, text=True)
            assert "Running" in result.stdout
            
            time.sleep(5.0)
            
            result = subprocess.run(["python3", "-m", "sari.mcp.cli", "daemon", "status", "--daemon-port", port], env=test_env, capture_output=True, text=True)
            assert "Stopped" in result.stdout
            
        finally:
            subprocess.run(["python3", "-m", "sari.mcp.cli", "daemon", "stop", "--daemon-port", port], env=test_env)


from typing import Dict, List, Set, Any
from dataclasses import dataclass, field

@dataclass(frozen=True)
class Profile:
    name: str
    extensions: Set[str] = field(default_factory=set)
    filenames: Set[str] = field(default_factory=set)
    globs: List[str] = field(default_factory=list)
    detect_files: List[str] = field(default_factory=list)

PROFILES: Dict[str, Profile] = {
    "core": Profile(
        name="core",
        extensions={".md", ".mdx", ".yaml", ".yml", ".json", ".toml", ".ini", ".cfg", ".conf", ".properties"},
        filenames={".env", "Makefile", "Dockerfile"},
        globs=[".env.*"]
    ),
    "web": Profile(
        name="web",
        extensions={".js", ".jsx", ".ts", ".tsx", ".html", ".css", ".scss", ".less", ".vue", ".svelte", ".astro", ".graphql", ".gql"},
        filenames={"package.json", "tsconfig.json", "webpack.config.js", "next.config.js"},
        detect_files=["package.json", "tsconfig.json", "vite.config.*", "webpack.config.*"]
    ),
    "python": Profile(
        name="python",
        extensions={".py", ".pyi", ".ipynb"},
        filenames={"pyproject.toml", "requirements.txt", "Pipfile", "setup.py"},
        detect_files=["pyproject.toml", "requirements.txt", "Pipfile", "setup.py"]
    ),
    "java": Profile(
        name="java",
        extensions={".java", ".kt", ".kts", ".gradle", ".xml"},
        filenames={"pom.xml", "build.gradle", "settings.gradle", "gradle.properties"},
        detect_files=["pom.xml", "build.gradle", "settings.gradle"]
    ),
    "go": Profile(
        name="go",
        extensions={".go"},
        filenames={"go.mod", "go.sum"},
        detect_files=["go.mod"]
    ),
    "rust": Profile(
        name="rust",
        extensions={".rs"},
        filenames={"Cargo.toml", "Cargo.lock"},
        detect_files=["Cargo.toml"]
    ),
    "cpp": Profile(
        name="cpp",
        extensions={".c", ".h", ".cpp", ".hpp", ".cc", ".cxx", ".hh", ".hxx"},
        filenames={"CMakeLists.txt"},
        detect_files=["CMakeLists.txt", "meson.build"]
    ),
    "infra": Profile(
        name="infra",
        extensions={".tf", ".tfvars", ".hcl", ".yaml", ".yml"},
        filenames={"Dockerfile", "docker-compose.yml"},
        detect_files=["Dockerfile", "docker-compose.yml", "*.tf", "*.tfvars"]
    ),
    # Add more as per ARCHITECTURE.md...
}



#!/usr/bin/env python3
"""
Sari CLI - Command-line interface for daemon management.

Usage:
    sari daemon start [-d]   Start daemon (foreground or daemonized)
    sari daemon stop         Stop running daemon
    sari daemon status       Check daemon status
    sari proxy               Run in proxy mode (stdio ↔ daemon)
"""
import argparse
import json
import os
import signal
import socket
import subprocess
import sys
import time
import urllib.parse
import urllib.request
import ipaddress
try:
    import psutil
except ImportError:
    psutil = None
from pathlib import Path
from typing import Optional, Dict, Any, Tuple

# Add project root to sys.path for absolute imports
SCRIPT_DIR = Path(__file__).parent
REPO_ROOT = SCRIPT_DIR.parent
if str(REPO_ROOT) not in sys.path:
    sys.path.insert(0, str(REPO_ROOT))


from sari.core.workspace import WorkspaceManager
from sari.core.server_registry import ServerRegistry, get_registry_path
from sari.core.config import Config
from sari.core.db import LocalSearchDB
from sari.core.daemon_resolver import resolve_daemon_address as get_daemon_address
from sari.mcp.tools.grep_and_read import execute_grep_and_read
from sari.mcp.tools.save_snippet import build_save_snippet
from sari.mcp.tools.get_snippet import build_get_snippet
from sari.mcp.tools.archive_context import build_archive_context
from sari.mcp.tools.get_context import build_get_context
from sari.mcp.tools.dry_run_diff import build_dry_run_diff


DEFAULT_HOST = "127.0.0.1"
DEFAULT_PORT = 47779
PID_FILE = WorkspaceManager.get_global_data_dir() / "daemon.pid"
DEFAULT_HTTP_HOST = "127.0.0.1"
DEFAULT_HTTP_PORT = 47777

def _arg(args: Any, name: str, default: Any = None) -> Any:
    return getattr(args, name, default)

def _pid_file_path() -> Path:
    return WorkspaceManager.get_global_data_dir() / "daemon.pid"


def _package_config_path() -> Path:
    return Path(__file__).parent.parent / "config" / "config.json"


def _load_config(workspace_root: str) -> Config:
    cfg_path = WorkspaceManager.resolve_config_path(workspace_root)
    return Config.load(cfg_path, workspace_root_override=workspace_root)


def _load_local_db(workspace_root: Optional[str] = None):
    root = workspace_root or WorkspaceManager.resolve_workspace_root()
    cfg_path = WorkspaceManager.resolve_config_path(root)
    cfg = Config.load(cfg_path, workspace_root_override=root)
    db = LocalSearchDB(cfg.db_path)
    return db, cfg.workspace_roots, root

def _load_server_info(workspace_root: str) -> Optional[dict]:
    """Legacy server.json location for backward compatibility."""
    server_json = Path(workspace_root) / ".codex" / "tools" / "sari" / "data" / "server.json"
    if not server_json.exists():
        return None
    try:
        return json.loads(server_json.read_text(encoding="utf-8"))
    except Exception:
        return None


def _load_registry_instances() -> Dict[str, Any]:
    try:
        reg_file = get_registry_path()
        if reg_file.exists():
            return json.loads(reg_file.read_text(encoding="utf-8")).get("instances", {})
    except Exception:
        pass
    return {}


def _is_loopback(host: str) -> bool:
    h = (host or "").strip().lower()
    if h == "localhost":
        return True
    try:
        return ipaddress.ip_address(h).is_loopback
    except ValueError:
        return False


def _enforce_loopback(host: str) -> None:
    # Security: Always enforce loopback. No overrides allowed.
    if not _is_loopback(host):
        raise RuntimeError(
            f"sari loopback-only: server_host must be 127.0.0.1/localhost/::1 (got={host}). "
            "Remote access is NOT supported for security."
        )


def _get_http_host_port(host_override: Optional[str] = None, port_override: Optional[int] = None) -> tuple[str, int]:
    """Get active HTTP server address with explicit/env/server/registry/config priority."""
    env_host = os.environ.get("SARI_HTTP_API_HOST") or os.environ.get("SARI_HTTP_HOST")
    env_port = os.environ.get("SARI_HTTP_API_PORT") or os.environ.get("SARI_HTTP_PORT")
    workspace_root = WorkspaceManager.resolve_workspace_root()
    cfg = _load_config(str(workspace_root))

    host = cfg.http_api_host or DEFAULT_HTTP_HOST
    port = int(cfg.http_api_port or DEFAULT_HTTP_PORT)

    try:
        ws_info = ServerRegistry().get_workspace(str(workspace_root))
        if ws_info:
            if ws_info.get("http_host"):
                host = str(ws_info.get("http_host"))
            if ws_info.get("http_port"):
                port = int(ws_info.get("http_port"))
    except Exception:
        pass

    server_info = _load_server_info(str(workspace_root))
    if server_info:
        try:
            if server_info.get("host"):
                host = str(server_info.get("host"))
            if server_info.get("port"):
                port = int(server_info.get("port"))
        except Exception:
            pass

    if env_host:
        host = env_host
    if env_port:
        try:
            port = int(env_port)
        except (TypeError, ValueError):
            pass

    if host_override:
        host = host_override
    if port_override is not None:
        port = int(port_override)

    return host, port



def _request_http(path: str, params: dict, host: Optional[str] = None, port: Optional[int] = None) -> dict:
    host, port = _get_http_host_port(host, port)
    _enforce_loopback(host)
    qs = urllib.parse.urlencode(params)
    url = f"http://{host}:{port}{path}?{qs}"
    with urllib.request.urlopen(url, timeout=3.0) as r:
        return json.loads(r.read().decode("utf-8"))


def _is_http_running(host: str, port: int, timeout: float = 0.4) -> bool:
    _enforce_loopback(host)
    try:
        url = f"http://{host}:{port}/health"
        with urllib.request.urlopen(url, timeout=timeout) as r:
            if r.status != 200:
                return False
            payload = json.loads(r.read().decode("utf-8"))
            return bool(payload.get("ok"))
    except Exception:
        return False


def _identify_sari_daemon(host: str, port: int, timeout: float = 1.0) -> Optional[Dict[str, Any]]:
    """Return identify payload if the server speaks Sari MCP."""
    try:
        with socket.create_connection((host, port), timeout=timeout) as sock:
            sock.settimeout(timeout)
            body = json.dumps({"jsonrpc": "2.0", "id": 1, "method": "sari/identify"}).encode("utf-8")
            header = f"Content-Length: {len(body)}\r\n\r\n".encode("ascii")
            sock.sendall(header + body)

            f = sock.makefile("rb")
            headers = {}
            while True:
                line = f.readline()
                if not line:
                    return None
                line = line.strip()
                if not line:
                    break
                if b":" in line:
                    k, v = line.split(b":", 1)
                    headers[k.strip().lower()] = v.strip()

            try:
                content_length = int(headers.get(b"content-length", b"0"))
            except ValueError:
                return None
            if content_length <= 0:
                return None
            resp_body = f.read(content_length)
            if not resp_body:
                return None
            resp = json.loads(resp_body.decode("utf-8"))

            result = resp.get("result") or {}
            if result.get("name") == "sari":
                return result
    except Exception:
        pass

    # Legacy fallback: probe "ping" and accept "Server not initialized" error
    try:
        with socket.create_connection((host, port), timeout=timeout) as sock:
            sock.settimeout(timeout)
            body = json.dumps({"jsonrpc": "2.0", "id": 1, "method": "ping"}).encode("utf-8")
            header = f"Content-Length: {len(body)}\r\n\r\n".encode("ascii")
            sock.sendall(header + body)

            f = sock.makefile("rb")
            headers = {}
            while True:
                line = f.readline()
                if not line:
                    return None
                line = line.strip()
                if not line:
                    break
                if b":" in line:
                    k, v = line.split(b":", 1)
                    headers[k.strip().lower()] = v.strip()

            try:
                content_length = int(headers.get(b"content-length", b"0"))
            except ValueError:
                return None
            if content_length <= 0:
                return None
            resp_body = f.read(content_length)
            if not resp_body:
                return None
            resp = json.loads(resp_body.decode("utf-8"))
            err = resp.get("error") or {}
            msg = (err.get("message") or "").lower()
            if "server not initialized" in msg:
                return {"name": "sari", "version": "legacy", "protocolVersion": ""}
    except Exception:
        pass

    return None


def _ensure_workspace_http(daemon_host: str, daemon_port: int, workspace_root: Optional[str] = None) -> bool:
    """Ensure workspace is initialized so HTTP server is started/registered."""
    try:
        root = workspace_root or os.environ.get("SARI_WORKSPACE_ROOT") or WorkspaceManager.resolve_workspace_root()
        with socket.create_connection((daemon_host, daemon_port), timeout=1.0) as sock:
            body = json.dumps({
                "jsonrpc": "2.0",
                "id": 1,
                "method": "initialize",
                "params": {"rootUri": f"file://{root}", "capabilities": {}},
            }).encode("utf-8")
            header = f"Content-Length: {len(body)}\r\n\r\n".encode("ascii")
            sock.sendall(header + body)
            f = sock.makefile("rb")
            headers = {}
            while True:
                line = f.readline()
                if not line:
                    break
                line = line.strip()
                if not line:
                    break
                if b":" in line:
                    k, v = line.split(b":", 1)
                    headers[k.strip().lower()] = v.strip()
            try:
                content_length = int(headers.get(b"content-length", b"0"))
            except ValueError:
                content_length = 0
            if content_length > 0:
                f.read(content_length)
        return True
    except Exception:
        return False


def _request_mcp_status(daemon_host: str, daemon_port: int, workspace_root: Optional[str] = None) -> Optional[dict]:
    try:
        root = workspace_root or os.environ.get("SARI_WORKSPACE_ROOT") or WorkspaceManager.resolve_workspace_root()
        with socket.create_connection((daemon_host, daemon_port), timeout=2.0) as sock:
            def _send(payload: dict) -> dict:
                body = json.dumps(payload).encode("utf-8")
                header = f"Content-Length: {len(body)}\r\n\r\n".encode("ascii")
                sock.sendall(header + body)
                f = sock.makefile("rb")
                headers = {}
                while True:
                    line = f.readline()
                    if not line:
                        break
                    line = line.strip()
                    if not line:
                        break
                    if b":" in line:
                        k, v = line.split(b":", 1)
                        headers[k.strip().lower()] = v.strip()
                try:
                    content_length = int(headers.get(b"content-length", b"0"))
                except ValueError:
                    content_length = 0
                body = f.read(content_length) if content_length > 0 else b""
                return json.loads(body.decode("utf-8")) if body else {}

            _send({
                "jsonrpc": "2.0",
                "id": 1,
                "method": "initialize",
                "params": {"rootUri": f"file://{root}", "capabilities": {}},
            })
            resp = _send({
                "jsonrpc": "2.0",
                "id": 2,
                "method": "tools/call",
                "params": {"name": "status", "arguments": {"details": True}},
            })
            return resp.get("result") or resp
    except Exception:
        return None


def _probe_sari_daemon(host: str, port: int, timeout: float = 0.3) -> bool:
    """Verify the server speaks Sari MCP (framed JSON-RPC)."""
    return _identify_sari_daemon(host, port, timeout=timeout) is not None


def _port_in_use(host: str, port: int) -> bool:
    try:
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            s.bind((host, port))
            return False
    except OSError:
        return True


def is_daemon_running(host: str, port: int) -> bool:
    """Check if a Sari daemon is running on the given port."""
    return _probe_sari_daemon(host, port)


def read_pid() -> Optional[int]:
    """Read PID from pidfile."""
    try:
        workspace_root = os.environ.get("SARI_WORKSPACE_ROOT") or WorkspaceManager.resolve_workspace_root()
        inst = ServerRegistry().resolve_workspace_daemon(str(workspace_root))
        if inst and inst.get("pid"):
            return int(inst.get("pid"))
    except Exception:
        pass
    for path in (_pid_file_path(), PID_FILE):
        if path.exists():
            try:
                return int(path.read_text().strip())
            except (ValueError, OSError):
                pass
    return None


def remove_pid() -> None:
    """Remove pidfile manually (if daemon crashed)."""
    for path in (_pid_file_path(), PID_FILE):
        try:
            if path.exists():
                path.unlink()
        except Exception:
            pass


def _get_local_version() -> str:
    try:
        from sari.version import __version__ as v
        return v or ""
    except Exception:
        return os.environ.get("SARI_VERSION", "") or ""


def _build_start_args(
    daemonize: bool = True,
    daemon_host: str = "",
    daemon_port: Optional[int] = None,
    http_host: str = "",
    http_port: Optional[int] = None,
) -> argparse.Namespace:
    return argparse.Namespace(
        daemonize=daemonize,
        daemon_host=daemon_host,
        daemon_port=daemon_port,
        http_host=http_host,
        http_port=http_port,
    )


def _start_daemon_background(
    daemon_host: str = "",
    daemon_port: Optional[int] = None,
    http_host: str = "",
    http_port: Optional[int] = None,
) -> bool:
    start_args = _build_start_args(
        daemonize=True,
        daemon_host=daemon_host,
        daemon_port=daemon_port,
        http_host=http_host,
        http_port=http_port,
    )
    return cmd_daemon_start(start_args) == 0


def _needs_upgrade_or_drain(identify: Optional[Dict[str, Any]]) -> bool:
    if not identify:
        return False
    existing_version = identify.get("version") or ""
    local_version = _get_local_version()
    draining = bool(identify.get("draining"))
    needs_upgrade = bool(existing_version and local_version and existing_version != local_version)
    return bool(needs_upgrade or draining)


def _ensure_daemon_running(
    daemon_host: str,
    daemon_port: int,
    http_host: str = "",
    http_port: Optional[int] = None,
    allow_upgrade: bool = False,
) -> Tuple[str, int, bool]:
    identify = _identify_sari_daemon(daemon_host, daemon_port)
    if identify and not (allow_upgrade and _needs_upgrade_or_drain(identify)):
        return daemon_host, daemon_port, True

    if allow_upgrade and identify and _needs_upgrade_or_drain(identify):
        _start_daemon_background(daemon_host, daemon_port, http_host, http_port)
        daemon_host, daemon_port = get_daemon_address()
        return daemon_host, daemon_port, is_daemon_running(daemon_host, daemon_port)

    if not is_daemon_running(daemon_host, daemon_port):
        _start_daemon_background(daemon_host, daemon_port, http_host, http_port)
        daemon_host, daemon_port = get_daemon_address()
        return daemon_host, daemon_port, is_daemon_running(daemon_host, daemon_port)

    return daemon_host, daemon_port, True


def cmd_daemon_start(args):
    """Start the daemon."""
    explicit_port = bool(_arg(args, "daemon_port"))
    force_start = (os.environ.get("SARI_DAEMON_FORCE_START") or "").strip().lower() in {"1", "true", "yes", "on"}
    workspace_root = os.environ.get("SARI_WORKSPACE_ROOT") or WorkspaceManager.resolve_workspace_root()
    registry = ServerRegistry()

    if _arg(args, "daemon_host") or _arg(args, "daemon_port"):
        host = _arg(args, "daemon_host") or DEFAULT_HOST
        port = int(_arg(args, "daemon_port") or DEFAULT_PORT)
    else:
        inst = registry.resolve_workspace_daemon(str(workspace_root))
        if inst and inst.get("port"):
            host = inst.get("host") or DEFAULT_HOST
            port = int(inst.get("port"))
        else:
            host, port = get_daemon_address()

    identify = _identify_sari_daemon(host, port)
    if identify:
        if explicit_port:
            ws_inst = registry.resolve_workspace_daemon(str(workspace_root))
            same_instance = bool(ws_inst and int(ws_inst.get("port", 0)) == int(port))
            if not same_instance:
                # Requested explicit port is occupied by another daemon instance.
                stop_args = argparse.Namespace(daemon_host=host, daemon_port=port)
                cmd_daemon_stop(stop_args)
                identify = _identify_sari_daemon(host, port)
                if identify:
                    print(f"❌ Port {port} is occupied by another running daemon.", file=sys.stderr)
                    return 1

        if not force_start and not _needs_upgrade_or_drain(identify):
            pid = read_pid()
            print(f"✅ Daemon already running on {host}:{port}")
            if not pid:
                print("⚠️  PID file missing. Another process may be using this port.")
                print("   Hint: Try a different port: SARI_DAEMON_PORT=47790 sari daemon start -d")
            return 0
        if explicit_port:
            print(f"❌ Port {port} is already in use by a running Sari daemon.", file=sys.stderr)
            return 1
        port = registry.find_free_port(start_port=47790)
        print(f"⚠️  Starting new daemon on free port {port} (upgrade/drain).")
    elif _port_in_use(host, port):
        if explicit_port:
            print(f"❌ Port {port} is already in use by another process.", file=sys.stderr)
            return 1
        # Choose a free port to avoid collisions with other MCP daemons.
        free_port = registry.find_free_port(start_port=47790)
        print(f"⚠️  Port {port} is in use. Switching to free port {free_port}.")
        port = free_port

    # Go up 3 levels: sari/mcp/cli.py -> sari/mcp -> sari/ -> (repo root)
    repo_root = Path(__file__).parent.parent.parent.resolve()
    env = os.environ.copy()
    env["PYTHONPATH"] = str(repo_root) + (os.pathsep + env["PYTHONPATH"] if env.get("PYTHONPATH") else "")
    env["SARI_DAEMON_AUTOSTART"] = "1"
    env["SARI_WORKSPACE_ROOT"] = workspace_root
    env["SARI_DAEMON_PORT"] = str(port)
    if _arg(args, "daemon_host"):
        env["SARI_DAEMON_HOST"] = _arg(args, "daemon_host")
    if _arg(args, "daemon_port"):
        env["SARI_DAEMON_PORT"] = str(_arg(args, "daemon_port"))
    if _arg(args, "http_host"):
        env["SARI_HTTP_API_HOST"] = _arg(args, "http_host")
    if _arg(args, "http_port") is not None:
        env["SARI_HTTP_API_PORT"] = str(_arg(args, "http_port"))

    if _arg(args, "daemonize"):
        # Start in background
        print(f"Starting daemon on {host}:{port} (background)...")

        proc = subprocess.Popen(
            [sys.executable, "-m", "sari.mcp.daemon"],
            cwd=repo_root.parent,
            start_new_session=True,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
            env=env,
        )

        # PID file will be written by the daemon process itself

        # Wait for startup
        for _ in range(30):
            if is_daemon_running(host, port):
                print(f"✅ Daemon started (PID: {proc.pid})")
                return 0
            time.sleep(0.1)

        print("❌ Daemon failed to start", file=sys.stderr)
        return 1
    else:
        # Start in foreground
        print(f"Starting daemon on {host}:{port} (foreground, Ctrl+C to stop)...")

        try:
            # Import and run directly
            os.environ["SARI_DAEMON_AUTOSTART"] = "1"
            os.environ["SARI_WORKSPACE_ROOT"] = workspace_root
            os.environ["PYTHONPATH"] = str(repo_root) + (os.pathsep + os.environ["PYTHONPATH"] if os.environ.get("PYTHONPATH") else "")
            if _arg(args, "daemon_host"):
                os.environ["SARI_DAEMON_HOST"] = _arg(args, "daemon_host")
            if _arg(args, "daemon_port"):
                os.environ["SARI_DAEMON_PORT"] = str(_arg(args, "daemon_port"))
            if _arg(args, "http_host"):
                os.environ["SARI_HTTP_API_HOST"] = _arg(args, "http_host")
            if _arg(args, "http_port") is not None:
                os.environ["SARI_HTTP_API_PORT"] = str(_arg(args, "http_port"))
            from sari.mcp.daemon import main as daemon_main
            import asyncio
            asyncio.run(daemon_main())
        except KeyboardInterrupt:
            print("\nDaemon stopped.")

        return 0


def cmd_daemon_stop(args):
    """Stop the daemon."""
    def _kill_pid_immediate(pid: int, label: str) -> bool:
        if not pid:
            return False
        try:
            if os.name == "nt":
                subprocess.run(["taskkill", "/F", "/T", "/PID", str(pid)], capture_output=True, check=False)
                print(f"Executed taskkill for {label} PID {pid}")
                return True
            # User-requested fast stop path: no long graceful wait.
            os.kill(pid, signal.SIGTERM)
            time.sleep(0.15)
            try:
                os.kill(pid, 0)
                os.kill(pid, signal.SIGKILL)
                print(f"Sent SIGKILL to {label} PID {pid}")
            except OSError:
                print(f"Stopped {label} PID {pid}")
            return True
        except ProcessLookupError:
            return True
        except PermissionError:
            print(f"Permission denied while stopping {label} PID {pid}")
            return False

    def _registry_targets(host: str, port: int, pid_hint: Optional[int]) -> Tuple[set[str], set[int]]:
        boot_ids: set[str] = set()
        http_pids: set[int] = set()
        try:
            reg = ServerRegistry()
            data = reg._load()  # internal load is acceptable for stop-path recovery
            daemons = data.get("daemons", {}) or {}
            workspaces = data.get("workspaces", {}) or {}
            for boot_id, info in daemons.items():
                if str(info.get("host") or DEFAULT_HOST) != str(host):
                    continue
                if int(info.get("port") or 0) != int(port):
                    continue
                if pid_hint and int(info.get("pid") or 0) not in {0, int(pid_hint)}:
                    continue
                boot_ids.add(str(boot_id))
            for ws_info in workspaces.values():
                if str(ws_info.get("boot_id") or "") not in boot_ids:
                    continue
                http_pid = int(ws_info.get("http_pid") or 0)
                if http_pid > 0:
                    http_pids.add(http_pid)
        except Exception:
            pass
        return boot_ids, http_pids

    if _arg(args, "daemon_host") or _arg(args, "daemon_port"):
        host = _arg(args, "daemon_host") or DEFAULT_HOST
        port = int(_arg(args, "daemon_port") or DEFAULT_PORT)
    else:
        host, port = get_daemon_address()

    if not is_daemon_running(host, port):
        print("Daemon is not running")
        remove_pid()
        return 0

    pid = read_pid()
    if not pid:
        try:
            # Fallback: derive daemon pid from registry when pid file is stale/missing.
            reg = ServerRegistry()
            data = reg._load()
            for info in (data.get("daemons") or {}).values():
                if str(info.get("host") or DEFAULT_HOST) == str(host) and int(info.get("port") or 0) == int(port):
                    pid = int(info.get("pid") or 0) or None
                    if pid:
                        break
        except Exception:
            pid = None

    boot_ids, http_pids = _registry_targets(host, port, pid)
    for http_pid in sorted(http_pids):
        _kill_pid_immediate(http_pid, "http")

    if pid:
        try:
            _kill_pid_immediate(pid, "daemon")
            for _ in range(10):
                if not is_daemon_running(host, port):
                    break
                time.sleep(0.1)
            reg = ServerRegistry()
            for boot_id in boot_ids:
                reg.unregister_daemon(boot_id)
            if is_daemon_running(host, port):
                print("⚠️  Daemon port still responds after stop attempt.")
            else:
                print("✅ Daemon stopped")
            remove_pid()
            return 0

        except (ProcessLookupError, PermissionError):
            print("PID not found or permission denied, daemon may have crashed or locked")
            remove_pid()
            return 0
    else:
        # No PID available: at least clean stale registry mappings for this endpoint.
        try:
            reg = ServerRegistry()
            for boot_id in boot_ids:
                reg.unregister_daemon(boot_id)
        except Exception:
            pass
        remove_pid()
        print("No PID file found. Registry cleaned for matching daemon endpoint.")
        return 0


def cmd_daemon_status(args):
    """Check daemon status."""
    if _arg(args, "daemon_host") or _arg(args, "daemon_port"):
        host = _arg(args, "daemon_host") or DEFAULT_HOST
        port = int(_arg(args, "daemon_port") or DEFAULT_PORT)
    else:
        host, port = get_daemon_address()

    running = is_daemon_running(host, port)
    pid = read_pid()

    print(f"Host: {host}")
    print(f"Port: {port}")
    print(f"Status: {'🟢 Running' if running else '⚫ Stopped'}")

    if pid:
        print(f"PID: {pid}")

    if running:
        # Try to get workspace info
        try:
            with socket.create_connection((host, port), timeout=1.0) as sock:
                # Send a status request
                import json
                request = json.dumps({
                    "jsonrpc": "2.0",
                    "id": 1,
                    "method": "tools/call",
                    "params": {"name": "status", "arguments": {"details": True}}
                }) + "\n"
                sock.sendall(request.encode())

                # We'd need an initialize first, so just skip detailed status for now
        except Exception:
            pass

    return 0 if running else 1


def cmd_proxy(args):
    """Run in proxy mode (for MCP stdio)."""
    from sari.mcp.proxy import main as proxy_main
    proxy_main()


def _tcp_blocked(err: OSError) -> bool:
    return getattr(err, "errno", None) in (1, 13)


def cmd_auto(args):
    """Try TCP daemon/proxy first, fallback to STDIO server."""
    host, port = get_daemon_address()

    # Fast path: if TCP is blocked by sandbox, skip daemon/proxy.
    try:
        with socket.create_connection((host, port), timeout=0.1):
            pass
    except OSError as e:
        if _tcp_blocked(e):
            from sari.mcp.server import main as server_main
            server_main()
            return 0
        # Connection refused etc. We'll try to start daemon below.

    identify = _identify_sari_daemon(host, port)
    if identify and not _needs_upgrade_or_drain(identify):
        return cmd_proxy(args)
    if identify and _needs_upgrade_or_drain(identify):
        host, port, running = _ensure_daemon_running(host, port, allow_upgrade=True)
        if running:
            return cmd_proxy(args)

    # Try to start daemon in background, then proxy.
    if not is_daemon_running(host, port):
        _start_daemon_background()
        for _ in range(30):
            try:
                host, port = get_daemon_address()
                if is_daemon_running(host, port):
                    break
            except OSError as e:
                if _tcp_blocked(e):
                    from sari.mcp.server import main as server_main
                    server_main()
                    return 0
            time.sleep(0.1)

    if is_daemon_running(host, port):
        return cmd_proxy(args)

    # Final fallback to STDIO server.
    from sari.mcp.server import main as server_main
    server_main()
    return 0


def cmd_status(args):
    """Query HTTP status endpoint."""
    try:
        if args.daemon_host or args.daemon_port:
            daemon_host = args.daemon_host or DEFAULT_HOST
            daemon_port = int(args.daemon_port or DEFAULT_PORT)
        else:
            daemon_host, daemon_port = get_daemon_address()
        daemon_running = is_daemon_running(daemon_host, daemon_port)
        host, port = _get_http_host_port(args.http_host, args.http_port)
        http_running = _is_http_running(host, port)

        if not http_running:
            if not daemon_running:
                daemon_host, daemon_port, daemon_running = _ensure_daemon_running(
                    daemon_host,
                    daemon_port,
                    http_host=args.http_host,
                    http_port=args.http_port,
                    allow_upgrade=False,
                )
            if daemon_running:
                for _ in range(5):
                    _ensure_workspace_http(daemon_host, daemon_port)
                    host, port = _get_http_host_port(args.http_host, args.http_port)
                    http_running = _is_http_running(host, port)
                    if http_running:
                        break
                    time.sleep(0.1)
            if not http_running and daemon_running:
                fallback = _request_mcp_status(daemon_host, daemon_port)
                if fallback:
                    print(json.dumps(fallback, ensure_ascii=False, indent=2))
                    return 0
            print("❌ Error: Sari services are not fully running.")
            print(f"   Daemon: {'🟢' if daemon_running else '⚫'} {daemon_host}:{daemon_port}")
            print(f"   HTTP:   {'🟢' if http_running else '⚫'} {host}:{port}")
            try:
                ws_root = WorkspaceManager.resolve_workspace_root()
                server_info = _load_server_info(str(ws_root))
                if server_info and server_info.get("port") and int(server_info.get("port")) != int(port):
                    note_host = server_info.get("host") or host
                    note_port = server_info.get("port")
                    print(f"   Note: server.json reports {note_host}:{note_port}")
            except Exception:
                pass
            print("   Hint: Run `sari daemon start -d` to start both, or `sari --http-api` to start HTTP only.")
            return 1

        data = _request_http("/status", {}, host, port)
        print(json.dumps(data, ensure_ascii=False, indent=2))
        return 0
    except Exception as e:
        print(f"❌ Error: Could not connect to Sari HTTP server.")
        print(f"   Details: {e}")
        print(f"   Hint: Make sure the Deamon is running for this workspace.")
        return 1


def cmd_search(args):
    """Query HTTP search endpoint."""
    params = {"q": args.query, "limit": args.limit}
    if args.repo:
        params["repo"] = args.repo
    data = _request_http("/search", params)
    print(json.dumps(data, ensure_ascii=False, indent=2))
    return 0


def cmd_doctor(args):
    from sari.mcp.tools.doctor import execute_doctor
    include_network = True
    include_db = True
    include_port = True
    include_disk = True
    if args.no_network:
        include_network = False
    if args.no_db:
        include_db = False
    if args.no_port:
        include_port = False
    if args.no_disk:
        include_disk = False
    if args.include_network:
        include_network = True
    if args.include_db:
        include_db = True
    if args.include_port:
        include_port = True
    if args.include_disk:
        include_disk = True

    payload = execute_doctor(
        {
            "auto_fix": bool(args.auto_fix),
            "auto_fix_rescan": bool(args.auto_fix_rescan),
            "include_network": include_network,
            "include_db": include_db,
            "include_port": include_port,
            "include_disk": include_disk,
            "min_disk_gb": float(args.min_disk_gb),
        }
    )
    text = payload.get("content", [{}])[0].get("text", "")
    try:
        parsed = json.loads(text)
        print(json.dumps(parsed, ensure_ascii=False, indent=2))
    except Exception:
        print(text)
    return 0


def cmd_init(args):
    """Initialize workspace with Sari config."""
    workspace_root = Path(args.workspace).expanduser().resolve() if args.workspace else Path(WorkspaceManager.resolve_workspace_root()).resolve()
    cfg_path = Path(WorkspaceManager.resolve_config_path(str(workspace_root)))
    data_dir = workspace_root / ".codex" / "tools" / "sari" / "data"

    cfg_path.parent.mkdir(parents=True, exist_ok=True)
    data_dir.mkdir(parents=True, exist_ok=True)

    from sari.core.config import Config
    default_cfg = Config.get_defaults(str(workspace_root))

    data = {}
    if cfg_path.exists() and not args.force:
        try:
            data = json.loads(cfg_path.read_text(encoding="utf-8")) or {}
        except Exception:
            data = {}

    roots = data.get("roots") or data.get("workspace_roots") or []
    if not isinstance(roots, list):
        roots = []
    if str(workspace_root) not in roots:
        roots.append(str(workspace_root))
    data["roots"] = roots
    data.setdefault("db_path", default_cfg["db_path"])
    data.setdefault("exclude_dirs", default_cfg["exclude_dirs"])

    cfg_path.write_text(json.dumps(data, indent=2), encoding="utf-8")
    print(f"✅ Updated Sari config: {cfg_path}")
    print(f"\n🚀 Workspace initialized successfully at {workspace_root}")
    return 0


def cmd_call_graph(args):
    db, roots, _ = _load_local_db(args.workspace)
    try:
        payload = build_call_graph(
            {
                "symbol": args.symbol,
                "symbol_id": args.symbol_id or "",
                "path": args.path or "",
                "depth": args.depth,
                "include_path": args.include_path or [],
                "exclude_path": args.exclude_path or [],
                "sort": args.sort,
            },
            db,
            roots,
        )
        if args.format == "tree":
            print(payload.get("tree", ""))
        else:
            print(json.dumps(payload, ensure_ascii=False, indent=2))
        return 0
    finally:
        db.close()


def cmd_save_snippet(args):
    db, roots, _ = _load_local_db(args.workspace)
    try:
        payload = build_save_snippet(
            {
                "path": args.path,
                "start_line": args.start_line,
                "end_line": args.end_line,
                "tag": args.tag,
                "note": args.note or "",
                "commit": args.commit or "",
            },
            db,
            roots,
            indexer=None,
        )
        print(json.dumps(payload, ensure_ascii=False, indent=2))
        return 0
    finally:
        db.close()


def cmd_get_snippet(args):
    db, roots, _ = _load_local_db(args.workspace)
    try:
        payload = build_get_snippet(
            {
                "tag": args.tag or "",
                "query": args.query or "",
                "limit": args.limit,
                "remap": (not args.no_remap),
                "history": args.history,
                "update": args.update,
                "diff_path": args.diff_path or "",
            },
            db,
            roots,
        )
        print(json.dumps(payload, ensure_ascii=False, indent=2))
        return 0
    finally:
        db.close()


def cmd_archive_context(args):
    db, roots, _ = _load_local_db(args.workspace)
    try:
        payload = build_archive_context(
            {
                "topic": args.topic,
                "content": args.content,
                "tags": args.tags or [],
                "related_files": args.related_files or [],
                "source": args.source or "",
                "valid_from": args.valid_from or "",
                "valid_until": args.valid_until or "",
                "deprecated": args.deprecated,
            },
            db,
            indexer=None,
        )
        print(json.dumps(payload, ensure_ascii=False, indent=2))
        return 0
    finally:
        db.close()


def cmd_get_context(args):
    db, roots, _ = _load_local_db(args.workspace)
    try:
        payload = build_get_context({"topic": args.topic or "", "query": args.query or "", "limit": args.limit, "as_of": args.as_of or ""}, db)
        print(json.dumps(payload, ensure_ascii=False, indent=2))
        return 0
    finally:
        db.close()


def cmd_dry_run_diff(args):
    db, roots, _ = _load_local_db(args.workspace)
    try:
        if args.lint:
            os.environ["SARI_DRYRUN_LINT"] = "1"
        payload = build_dry_run_diff({"path": args.path, "content": args.content}, db, roots)
        print(json.dumps(payload, ensure_ascii=False, indent=2))
        return 0
    finally:
        db.close()

def cmd_grep_and_read(args):
    db, roots, _ = _load_local_db(args.workspace)
    try:
        payload = execute_grep_and_read(
            {
                "query": args.query,
                "repo": args.repo or "",
                "limit": args.limit,
                "read_limit": args.read_limit,
                "file_types": args.file_types or [],
                "path_pattern": args.path_pattern or "",
                "exclude_patterns": args.exclude_patterns or [],
                "recency_boost": bool(args.recency_boost),
                "use_regex": bool(args.use_regex),
                "case_sensitive": bool(args.case_sensitive),
                "context_lines": args.context_lines,
                "total_mode": args.total_mode or "exact",
            },
            db,
            roots,
        )
        text = payload.get("content", [{}])[0].get("text", "")
        try:
            parsed = json.loads(text)
            print(json.dumps(parsed, ensure_ascii=False, indent=2))
        except Exception:
            print(text)
        return 0
    finally:
        db.close()



def cmd_prune(args):
    """Prune old data from auxiliary tables."""
    db, roots, _ = _load_local_db(args.workspace)
    try:
        tables = [args.table] if args.table else ["snippets", "failed_tasks", "contexts"]
        total_pruned = 0
        print(f"🧹 Pruning data older than {args.days or '(default)'} days...")
        
        for table in tables:
            ttl = args.days
            if ttl is None:
                # Fallback to settings
                if table == "snippets": ttl = db.settings.STORAGE_TTL_DAYS_SNIPPETS
                elif table == "failed_tasks": ttl = db.settings.STORAGE_TTL_DAYS_FAILED_TASKS
                elif table == "contexts": ttl = db.settings.STORAGE_TTL_DAYS_CONTEXTS
                else: ttl = 30 # Safe fallback
            
            count = db.prune_data(table, ttl)
            if count > 0:
                print(f"   - {table}: Removed {count} records (older than {ttl} days)")
                total_pruned += count
        
        if total_pruned == 0:
            print("✨ No data to prune.")
        else:
            print(f"✅ Total pruned: {total_pruned} records")
        return 0
    finally:
        db.close()


def main():

    epilog = "\n".join([
        "Examples:",
        "  sari daemon start -d",
        "  sari status",
        "  sari doctor --auto-fix",
        "  sari search \"query\" --limit 10",
        "  sari call-graph --symbol process_file --depth 2",
        "  sari save-snippet --path src/app.py --start-line 10 --end-line 20 --tag db-idiom",
        "  sari get-snippet --tag db-idiom",
        "  sari archive-context --topic PricingLogic --content \"...\"",
        "  sari get-context --query PricingLogic",
        "  sari dry-run-diff --path src/app.py --content \"<new content>\"",
        "  sari auto",
    ])
    parser = argparse.ArgumentParser(
        prog="sari",
        description="Sari - Local Search MCP Server",
        formatter_class=argparse.RawTextHelpFormatter,
        epilog=epilog,
    )

    subparsers = parser.add_subparsers(dest="command", help="Commands")

    # daemon subcommand
    daemon_parser = subparsers.add_parser("daemon", help="Daemon management")
    daemon_sub = daemon_parser.add_subparsers(dest="daemon_command")

    # daemon start
    start_parser = daemon_sub.add_parser("start", help="Start daemon")
    start_parser.add_argument("-d", "--daemonize", action="store_true",
                              help="Run in background")
    start_parser.add_argument("--daemon-host", default="", help="Daemon host override (default: 127.0.0.1)")
    start_parser.add_argument("--daemon-port", type=int, default=None, help="Daemon port override (default: 47779)")
    start_parser.add_argument("--http-host", default="", help="HTTP host override (default: 127.0.0.1)")
    start_parser.add_argument("--http-port", type=int, default=None, help="HTTP port override (default: 47777)")
    start_parser.set_defaults(func=cmd_daemon_start)

    # daemon stop
    stop_parser = daemon_sub.add_parser("stop", help="Stop daemon")
    stop_parser.add_argument("--daemon-host", default="", help="Daemon host override")
    stop_parser.add_argument("--daemon-port", type=int, default=None, help="Daemon port override")
    stop_parser.set_defaults(func=cmd_daemon_stop)

    # daemon status
    status_parser = daemon_sub.add_parser("status", help="Check status")
    status_parser.add_argument("--daemon-host", default="", help="Daemon host override")
    status_parser.add_argument("--daemon-port", type=int, default=None, help="Daemon port override")
    status_parser.set_defaults(func=cmd_daemon_status)

    # proxy subcommand
    proxy_parser = subparsers.add_parser("proxy", help="Run in proxy mode")
    proxy_parser.set_defaults(func=cmd_proxy)

    # auto subcommand
    auto_parser = subparsers.add_parser("auto", help="Auto mode (daemon/proxy fallback)")
    auto_parser.set_defaults(func=cmd_auto)

    # status subcommand (HTTP)
    status_parser = subparsers.add_parser("status", help="Query HTTP status")
    status_parser.add_argument("--daemon-host", default="", help="Daemon host override (default: 127.0.0.1)")
    status_parser.add_argument("--daemon-port", type=int, default=None, help="Daemon port override (default: 47779)")
    status_parser.add_argument("--http-host", default="", help="HTTP host override (default: 127.0.0.1)")
    status_parser.add_argument("--http-port", type=int, default=None, help="HTTP port override (default: 47777)")
    status_parser.set_defaults(func=cmd_status)

    # doctor subcommand
    doctor_parser = subparsers.add_parser("doctor", help="Run diagnostics")
    doctor_parser.add_argument("--auto-fix", action="store_true", help="Attempt automatic fixes when possible")
    doctor_parser.add_argument("--auto-fix-rescan", action="store_true", help="Run scan_once after auto-fix")
    doctor_parser.add_argument("--include-network", action="store_true", help="Include network check")
    doctor_parser.add_argument("--no-network", action="store_true", help="Skip network check")
    doctor_parser.add_argument("--include-db", action="store_true", help="Include DB check")
    doctor_parser.add_argument("--no-db", action="store_true", help="Skip DB check")
    doctor_parser.add_argument("--include-port", action="store_true", help="Include port check")
    doctor_parser.add_argument("--no-port", action="store_true", help="Skip port check")
    doctor_parser.add_argument("--include-disk", action="store_true", help="Include disk check")
    doctor_parser.add_argument("--no-disk", action="store_true", help="Skip disk check")
    doctor_parser.add_argument("--min-disk-gb", type=float, default=1.0, help="Minimum disk space GB")
    doctor_parser.set_defaults(func=cmd_doctor)

    # search subcommand (HTTP)
    search_parser = subparsers.add_parser("search", help="Search via HTTP server")
    search_parser.add_argument("query", help="Search query")
    search_parser.add_argument("--repo", default="", help="Limit search to repo")
    search_parser.add_argument("--limit", type=int, default=10, help="Max results (default: 10)")
    search_parser.set_defaults(func=cmd_search)

    # init subcommand
    init_parser = subparsers.add_parser("init", help="Initialize workspace config")
    init_parser.add_argument("--workspace", default="", help="Workspace root (default: auto-detect)")
    init_parser.add_argument("--force", action="store_true", help="Overwrite existing config")
    init_parser.set_defaults(func=cmd_init)

    # call-graph
    cg_parser = subparsers.add_parser("call-graph", help="Call graph for a symbol")
    cg_parser.add_argument("--symbol", required=True, help="Target symbol name")
    cg_parser.add_argument("--symbol-id", default="", help="Optional symbol_id to disambiguate")
    cg_parser.add_argument("--path", default="", help="Optional db-path or file path to disambiguate")
    cg_parser.add_argument("--depth", type=int, default=2, help="Graph depth (default: 2)")
    cg_parser.add_argument("--format", default="json", choices=["json", "tree"], help="Output format")
    cg_parser.add_argument("--include-path", nargs="*", default=[], help="Include paths (prefix or substring)")
    cg_parser.add_argument("--exclude-path", nargs="*", default=[], help="Exclude paths (prefix or substring)")
    cg_parser.add_argument("--sort", default="line", choices=["line", "name"], help="Tree sort order")
    cg_parser.add_argument("--workspace", default="", help="Workspace root (default: auto-detect)")
    cg_parser.set_defaults(func=cmd_call_graph)

    # save-snippet
    ss_parser = subparsers.add_parser("save-snippet", help="Save code snippet")
    ss_parser.add_argument("--path", required=True, help="Path or path:start-end")
    ss_parser.add_argument("--start-line", type=int, default=None)
    ss_parser.add_argument("--end-line", type=int, default=None)
    ss_parser.add_argument("--tag", required=True, help="Tag for snippet")
    ss_parser.add_argument("--note", default="", help="Optional note")
    ss_parser.add_argument("--commit", default="", help="Optional commit hash")
    ss_parser.add_argument("--workspace", default="", help="Workspace root (default: auto-detect)")
    ss_parser.set_defaults(func=cmd_save_snippet)

    # get-snippet
    gs_parser = subparsers.add_parser("get-snippet", help="Get saved snippet")
    gs_parser.add_argument("--tag", default="", help="Tag to lookup")
    gs_parser.add_argument("--query", default="", help="Search query")
    gs_parser.add_argument("--limit", type=int, default=20, help="Max results")
    gs_parser.add_argument("--no-remap", action="store_true", help="Disable remap against current file")
    gs_parser.add_argument("--history", action="store_true", help="Include snippet versions")
    gs_parser.add_argument("--update", action="store_true", help="Write back remapped location/content")
    gs_parser.add_argument("--diff-path", default="", help="Write remap diff to file when --update is used")
    gs_parser.add_argument("--workspace", default="", help="Workspace root (default: auto-detect)")
    gs_parser.set_defaults(func=cmd_get_snippet)

    # archive-context
    ac_parser = subparsers.add_parser("archive-context", help="Archive domain context")
    ac_parser.add_argument("--topic", required=True, help="Context topic")
    ac_parser.add_argument("--content", required=True, help="Context content")
    ac_parser.add_argument("--tags", nargs="*", default=[], help="Tags")
    ac_parser.add_argument("--related-files", nargs="*", default=[], help="Related files")
    ac_parser.add_argument("--source", default="", help="Context source (doc/issue/link)")
    ac_parser.add_argument("--valid-from", default="", help="Valid from (timestamp or ISO date)")
    ac_parser.add_argument("--valid-until", default="", help="Valid until (timestamp or ISO date)")
    ac_parser.add_argument("--deprecated", action="store_true", help="Mark context as deprecated")
    ac_parser.add_argument("--workspace", default="", help="Workspace root (default: auto-detect)")
    ac_parser.set_defaults(func=cmd_archive_context)

    # get-context
    gc_parser = subparsers.add_parser("get-context", help="Get archived context")
    gc_parser.add_argument("--topic", default="", help="Topic to lookup")
    gc_parser.add_argument("--query", default="", help="Search query")
    gc_parser.add_argument("--limit", type=int, default=20, help="Max results")
    gc_parser.add_argument("--as-of", default="", help="Filter by validity timestamp or ISO date")
    gc_parser.add_argument("--workspace", default="", help="Workspace root (default: auto-detect)")
    gc_parser.set_defaults(func=cmd_get_context)

    # dry-run-diff
    dr_parser = subparsers.add_parser("dry-run-diff", help="Preview diff and syntax check")
    dr_parser.add_argument("--path", required=True, help="Path of file to edit")
    dr_parser.add_argument("--content", required=True, help="Proposed full file content")
    dr_parser.add_argument("--lint", action="store_true", help="Run lint if available")
    dr_parser.add_argument("--workspace", default="", help="Workspace root (default: auto-detect)")
    dr_parser.set_defaults(func=cmd_dry_run_diff)

    # grep-and-read
    gar_parser = subparsers.add_parser("grep-and-read", help="Search then read top files")
    gar_parser.add_argument("query", help="Search query")
    gar_parser.add_argument("--repo", default="", help="Limit search to repo")
    gar_parser.add_argument("--limit", type=int, default=8, help="Max results (default: 8)")
    gar_parser.add_argument("--read-limit", type=int, default=3, help="Files to read (default: 3)")
    gar_parser.add_argument("--file-types", nargs="*", default=[], help="Filter by file extension")
    gar_parser.add_argument("--path-pattern", default="", help="Glob pattern for path matching")
    gar_parser.add_argument("--exclude-patterns", nargs="*", default=[], help="Patterns to exclude")
    gar_parser.add_argument("--recency-boost", action="store_true", help="Boost recently modified files")
    gar_parser.add_argument("--use-regex", action="store_true", help="Treat query as regex")
    gar_parser.add_argument("--case-sensitive", action="store_true", help="Case-sensitive search")
    gar_parser.add_argument("--context-lines", type=int, default=5, help="Snippet lines (default: 5)")
    gar_parser.add_argument("--total-mode", default="exact", choices=["exact", "approx"], help="Total count mode")
    gar_parser.add_argument("--workspace", default="", help="Workspace root (default: auto-detect)")
    gar_parser.set_defaults(func=cmd_grep_and_read)

    # prune
    prune_parser = subparsers.add_parser("prune", help="Prune old data (maintenance)")
    prune_parser.add_argument("--days", type=int, default=None, help="Override TTL days (default: use settings)")
    prune_parser.add_argument("--table", choices=["snippets", "failed_tasks", "contexts"], help="Target specific table")
    prune_parser.add_argument("--workspace", default="", help="Workspace root (default: auto-detect)")
    prune_parser.set_defaults(func=cmd_prune)

    args = parser.parse_args()

    if not args.command:
        parser.print_help()
        return 1

    if args.command == "daemon" and not args.daemon_command:
        daemon_parser.print_help()
        return 1

    if hasattr(args, "func"):
        return args.func(args)

    return 0


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python3
"""
Repo candidates tool for Sari MCP Server.
"""
import json
from typing import Any, Dict, List
from sari.mcp.tools._util import mcp_response, pack_header, pack_line, pack_encode_id, pack_encode_text, pack_error, ErrorCode, resolve_root_ids

from sari.core.db import LocalSearchDB
from sari.mcp.telemetry import TelemetryLogger


def execute_repo_candidates(args: Dict[str, Any], db: LocalSearchDB, logger: TelemetryLogger = None, roots: List[str] = None) -> Dict[str, Any]:
    """Execute repo_candidates tool."""
    query = args.get("query", "")
    try:
        limit_arg = min(int(args.get("limit", 3)), 5)
    except (ValueError, TypeError):
        limit_arg = 3

    if not query.strip():
        return mcp_response(
            "repo_candidates",
            lambda: pack_error("repo_candidates", ErrorCode.INVALID_ARGS, "query is required"),
            lambda: {"error": {"code": ErrorCode.INVALID_ARGS.value, "message": "query is required"}, "isError": True},
        )

    def get_candidates():
        root_ids = resolve_root_ids(list(roots or []))
        candidates = db.repo_candidates(q=query, limit=limit_arg, root_ids=root_ids)
        for candidate in candidates:
            score = candidate.get("score", 0)
            if score >= 10:
                reason = f"High match ({score} files contain '{query}')"
            elif score >= 5:
                reason = f"Moderate match ({score} files)"
            else:
                reason = f"Low match ({score} files)"
            candidate["reason"] = reason
        return candidates

    # --- JSON Builder ---
    def build_json() -> Dict[str, Any]:
        candidates = get_candidates()
        return {
            "query": query,
            "candidates": candidates,
            "hint": "Use 'repo' parameter in search to narrow down scope after selection",
        }

    # --- PACK1 Builder ---
    def build_pack() -> str:
        candidates = get_candidates()

        # Header
        kv = {"q": pack_encode_text(query), "limit": limit_arg}
        lines = [
            pack_header("repo_candidates", kv, returned=len(candidates))
        ]

        # Records
        for c in candidates:
            # r:repo=<repo> score=<score> reason=<reason>
            kv_line = {
                "repo": pack_encode_id(c["repo"]),
                "score": str(c["score"]),
                "reason": pack_encode_text(c["reason"])
            }
            lines.append(pack_line("r", kv_line))

        return "\n".join(lines)

    if logger:
        # We need candidate count for logging, but don't want to run query twice optimally.
        # But for simplicity in this structure, we let builders run query.
        # Telemetry here might be slightly off if we don't capture result from mcp_response,
        # but execute_repo_candidates returns the result, so we can't easily hook in unless we move logging inside builders or after mcp_response.
        # Let's log *after* mcp_response call by peeking result, or just log query intent.
        logger.log_telemetry(f"tool=repo_candidates query='{query}' limit={limit_arg}")

    return mcp_response("repo_candidates", build_pack, build_json)


import pytest
import os
import sys
import asyncio
from unittest.mock import MagicMock, patch

# Mock dependencies
sys.modules['psutil'] = MagicMock()
sys.modules['watchdog'] = MagicMock()
sys.modules['watchdog.observers'] = MagicMock()
sys.modules['watchdog.events'] = MagicMock()

def test_coverage_final_push():
    # 1. sari.main
    import sari.main
    with patch('sys.argv', ['sari', '--help']):
        try: sari.main.main()
        except SystemExit: pass

    # 2. sari.mcp.daemon
    from sari.mcp.daemon import SariDaemon
    daemon = SariDaemon("127.0.0.1", 47779)
    assert daemon.host == "127.0.0.1"

    # 3. sari.mcp.session
    from sari.mcp.session import Session
    s = Session(MagicMock(), MagicMock())
    assert s.running is True

    # 4. sari.core.main
    import sari.core.main
    # If there's a main class or function, call it
    if hasattr(sari.core.main, 'LocalSearchCore'):
        c = sari.core.main.LocalSearchCore(MagicMock())

    # 5. sari.mcp.tools.call_graph
    from sari.mcp.tools.call_graph import execute_call_graph
    db = MagicMock()
    db.search_symbols.return_value = []
    execute_call_graph({"symbol": "test"}, db, ["/tmp"])

    # 6. sari.core.cjk
    from sari.core.cjk import lindera_available
    lindera_available()

    # 7. sari.core.db.schema
    from sari.core.db.schema import init_schema
    init_schema(MagicMock())

def test_indexer_advanced(tmp_path):
    from sari.core.indexer.main import Indexer
    cfg = MagicMock()
    cfg.workspace_roots = [str(tmp_path)]
    cfg.max_depth = 30
    
    mock_settings = MagicMock()
    mock_settings.MAX_DEPTH = 30
    mock_settings.INDEX_MEM_MB = 1024
    mock_settings.INDEX_WORKERS = 2
    mock_settings.get_int.side_effect = lambda key, default: default
    cfg.settings = mock_settings
    
    db = MagicMock()
    with patch('sari.core.db.storage.GlobalStorageManager.get_instance'):
        indexer = Indexer(cfg, db, settings_obj=mock_settings)
        indexer.scan_once()
        # Call some internal methods for coverage
        indexer._retry_failed_tasks()
        indexer.get_queue_depths()

def test_session_async():
    from sari.mcp.session import Session
    reader = MagicMock()
    writer = MagicMock()
    # Mock reader.readline to return empty (EOF)
    reader.readline = MagicMock(return_value=asyncio.Future())
    reader.readline.return_value.set_result(b"")
    
    s = Session(reader, writer)
    loop = asyncio.new_event_loop()
    loop.run_until_complete(s.handle_connection())
    loop.close()
    assert s.running is False



import os
import sys
import hashlib
import json
import shutil
from pathlib import Path
from typing import Optional, List
from sari.core.settings import settings

class WorkspaceManager:
    """Manages workspace detection, boundary markers (.sariroot), and global paths."""
    settings = settings

    @staticmethod
    def set_settings(settings_obj):
        WorkspaceManager.settings = settings_obj

    @staticmethod
    def find_project_root(path: str) -> str:
        """Find the nearest directory containing .sariroot or .sari."""
        curr = Path(path).resolve()
        for parent in [curr] + list(curr.parents):
            if (parent / ".sariroot").exists() or (parent / ".sari").exists():
                return str(parent)
        return str(curr)

    @staticmethod
    def _strip_trailing_sep(path: str) -> str:
        if not path:
            return path
        stripped = path.rstrip(os.sep)
        # Keep filesystem root stable ("/" on POSIX).
        if not stripped:
            return os.sep
        return stripped

    @staticmethod
    def normalize_path(p: str) -> str:
        """Standard normalization for all workspace paths in Sari. Fixes Mac symlink issues."""
        if not p:
            p = os.getcwd()
        try:
            # Mac specific: ensure /tmp -> /private/tmp resolution for FK stability
            resolved = str(Path(p).expanduser().resolve())
            return WorkspaceManager._strip_trailing_sep(resolved)
        except Exception:
            # Fallback to absolute if resolve fails
            res = os.path.abspath(os.path.expanduser(p))
            if os.name == "nt": res = res.lower()
            return WorkspaceManager._strip_trailing_sep(res)

    @staticmethod
    def root_id(path: str) -> str:
        """Stable root id derived from project root boundary."""
        project_root = WorkspaceManager.find_project_root(path)
        digest = hashlib.sha1(project_root.encode("utf-8")).hexdigest()[:8]
        return f"root-{digest}"

    @staticmethod
    def root_id_for_workspace(workspace_root: str) -> str:
        """Stable root id for an explicit workspace root (no parent auto-lift)."""
        normalized = WorkspaceManager.normalize_path(workspace_root)
        digest = hashlib.sha1(normalized.encode("utf-8")).hexdigest()[:8]
        return f"root-{digest}"

    @staticmethod
    def _normalize_path(path: str, follow_symlinks: bool) -> str:
        expanded = os.path.expanduser(path)
        normalized = os.path.realpath(expanded) if follow_symlinks else os.path.abspath(expanded)
        if os.name == "nt": normalized = normalized.lower()
        return WorkspaceManager._strip_trailing_sep(normalized)

    @staticmethod
    def resolve_workspace_roots(root_uri: Optional[str] = None, config_roots: Optional[List[str]] = None) -> List[str]:
        roots = []
        if config_roots:
            for r in config_roots:
                if r: roots.append(WorkspaceManager.normalize_path(r))
        if root_uri:
            p = root_uri[7:] if root_uri.startswith("file://") else root_uri
            roots.append(WorkspaceManager.normalize_path(p))
        if WorkspaceManager.settings.WORKSPACE_ROOT:
            roots.append(WorkspaceManager.normalize_path(WorkspaceManager.settings.WORKSPACE_ROOT))
        return list(dict.fromkeys(roots)) or [WorkspaceManager.normalize_path(os.getcwd())]

    @staticmethod
    def resolve_workspace_root(root_uri: Optional[str] = None) -> str:
        """Resolve a single workspace root (primary)."""
        roots = WorkspaceManager.resolve_workspace_roots(root_uri=root_uri)
        root = roots[0] if roots else os.getcwd()
        return WorkspaceManager.normalize_path(root)

    @staticmethod
    def resolve_config_path(_repo_root: str) -> str:
        """
        Resolve config path.
        Priority:
        1) SARI_CONFIG (explicit override)
        2) workspace-local .sari/mcp-config.json (preferred)
        3) global ~/.config/sari/config.json (legacy fallback)
        """
        val = WorkspaceManager.settings.CONFIG_PATH
        if val:
            return str(Path(os.path.expanduser(val)).resolve())
        repo_root = _repo_root or os.getcwd()
        ws_root = WorkspaceManager.find_project_root(repo_root)
        preferred = WorkspaceManager.workspace_config_path(ws_root)
        WorkspaceManager._migrate_legacy_workspace_config(ws_root, preferred)
        if preferred.exists():
            return str(preferred)
        return str(Path(WorkspaceManager.settings.GLOBAL_CONFIG_DIR) / "config.json")

    @staticmethod
    def workspace_config_path(root_path: str) -> Path:
        """Preferred workspace-local config path."""
        root = Path(root_path)
        return root / WorkspaceManager.settings.WORKSPACE_CONFIG_DIR_NAME / "mcp-config.json"

    @staticmethod
    def legacy_workspace_config_path(root_path: str) -> Path:
        """Legacy workspace-local config path kept for backward compatibility."""
        root = Path(root_path)
        return root / WorkspaceManager.settings.WORKSPACE_CONFIG_DIR_NAME / "config.json"

    @staticmethod
    def _looks_like_sqlite(path: Path) -> bool:
        try:
            with path.open("rb") as f:
                head = f.read(16)
            return head.startswith(b"SQLite format 3")
        except Exception:
            return False

    @staticmethod
    def _migrate_legacy_workspace_config(root_path: str, preferred_path: Path) -> None:
        """
        One-way migration:
        - If legacy config.json is valid JSON and preferred file is absent, copy it.
        - Never migrate if legacy file looks like SQLite (corrupted/misused path).
        """
        legacy = WorkspaceManager.legacy_workspace_config_path(root_path)
        if preferred_path.exists() or not legacy.exists():
            return
        if WorkspaceManager._looks_like_sqlite(legacy):
            return
        try:
            data = json.loads(legacy.read_text(encoding="utf-8"))
            if isinstance(data, dict):
                preferred_path.parent.mkdir(parents=True, exist_ok=True)
                shutil.copy2(legacy, preferred_path)
        except Exception:
            return

    @staticmethod
    def get_global_data_dir() -> Path:
        return Path.home() / ".local" / "share" / "sari"

    @staticmethod
    def get_global_db_path() -> Path:
        return WorkspaceManager.get_global_data_dir() / "index.db"

    @staticmethod
    def get_workspace_data_dir(root_path: str) -> Path:
        root = Path(WorkspaceManager.normalize_path(root_path))
        return root / WorkspaceManager.settings.WORKSPACE_CONFIG_DIR_NAME

    @staticmethod
    def get_workspace_db_path(root_path: str) -> Path:
        return WorkspaceManager.get_workspace_data_dir(root_path) / "index.db"

    @staticmethod
    def get_global_log_dir() -> Path:
        return Path.home() / "Library" / "Logs" / "sari" if sys.platform == "darwin" else WorkspaceManager.get_global_data_dir() / "logs"

    @staticmethod
    def get_engine_index_dir(policy: Optional[str] = None, roots: Optional[List[str]] = None, root_id: Optional[str] = None) -> Path:
        """
        Index directory selector.
        Policy:
        - global (default): single index
        - roots_hash: per-roots group index (legacy-friendly)
        - per_root: one index per root_id
        """
        policy = (policy or WorkspaceManager.settings.ENGINE_INDEX_POLICY or "global").lower()
        base = WorkspaceManager.get_global_data_dir() / "index"
        if policy in {"global", "single"}:
            return base / "global"
        if policy in {"roots_hash", "legacy"}:
            roots = roots or []
            seed = "::".join(sorted([WorkspaceManager._normalize_path(r, WorkspaceManager.settings.FOLLOW_SYMLINKS) for r in roots]))
            digest = hashlib.sha1(seed.encode("utf-8")).hexdigest()[:8] if seed else "default"
            return base / f"roots-{digest}"
        if policy in {"per_root", "shard"}:
            rid = root_id or (WorkspaceManager.root_id_for_workspace(roots[0]) if roots else "root-default")
            return base / f"{rid}"
        return base / "global"

    @staticmethod
    def ensure_sari_dir(root_path: str) -> None:
        root = Path(root_path)
        sari_dir = root / WorkspaceManager.settings.WORKSPACE_CONFIG_DIR_NAME
        sari_dir.mkdir(parents=True, exist_ok=True)
        gitignore = root / ".gitignore"
        entry = f"{WorkspaceManager.settings.WORKSPACE_CONFIG_DIR_NAME}/"
        try:
            if not gitignore.exists():
                gitignore.write_text(entry + "\n", encoding="utf-8")
                return
            text = gitignore.read_text(encoding="utf-8")
            if entry not in text:
                with gitignore.open("a", encoding="utf-8") as f:
                    if not text.endswith("\n"):
                        f.write("\n")
                    f.write(entry + "\n")
        except Exception:
            pass



import pytest
import argparse
import sys
from unittest.mock import MagicMock, patch

# Mock psutil
sys.modules['psutil'] = MagicMock()

from sari.mcp.cli import cmd_doctor, cmd_search, cmd_daemon_start, cmd_daemon_stop

def test_cmd_doctor():
    args = argparse.Namespace(
        auto_fix=True, auto_fix_rescan=False, 
        no_network=False, no_db=False, no_port=False, no_disk=False,
        include_network=True, include_db=True, include_port=True, include_disk=True,
        min_disk_gb=1.0
    )
    with patch('sari.mcp.tools.doctor.execute_doctor') as mock_exec:
        mock_exec.return_value = {"content": [{"text": "{}"}]}
        ret = cmd_doctor(args)
        assert ret == 0

def test_cmd_search():
    args = argparse.Namespace(query="test", limit=10, repo=None)
    with patch('sari.mcp.cli._request_http', return_value={"results": []}):
        with patch('sari.mcp.cli._get_http_host_port', return_value=("127.0.0.1", 47777)):
            ret = cmd_search(args)
            assert ret == 0

def test_cmd_daemon_stop():
    args = argparse.Namespace()
    with patch('sari.mcp.cli.read_pid', return_value=1234):
        with patch('os.kill') as mock_kill:
            with patch('sari.mcp.cli.ServerRegistry') as mock_registry_cls:
                mock_registry = MagicMock()
                mock_registry._load.return_value = {
                    "daemons": {"b1": {"host": "127.0.0.1", "port": 47779, "pid": 1234}},
                    "workspaces": {"/tmp/ws": {"boot_id": "b1", "http_pid": 4321}},
                }
                mock_registry_cls.return_value = mock_registry
                # First call: daemon is running, later calls: stopped
                with patch('sari.mcp.cli.is_daemon_running', side_effect=[True, False, False]):
                    ret = cmd_daemon_stop(args)
                    assert ret == 0
                    # daemon + http kill path should execute
                    assert mock_kill.called

def test_uninstall():
    from sari.uninstall import main as uninstall_main
    with patch('sys.argv', ['sari-uninstall']):
        with patch('builtins.input', return_value='y'):
            with patch('shutil.rmtree') as mock_rm:
                try:
                    uninstall_main()
                except SystemExit: pass
                assert mock_rm.called



import json
import os
import time
import socket
from contextlib import contextmanager
from pathlib import Path
from typing import Dict, Optional, Any, Iterable

# Cross-platform file locking
IS_WINDOWS = os.name == "nt"
if not IS_WINDOWS:
    import fcntl

def share_path() -> str:
    return os.path.join("share", "sari")

# Backward compatibility for legacy tests/tools that patch module-level path.
REGISTRY_FILE = Path.home() / ".local" / share_path() / "server.json"

def get_registry_path() -> Path:
    """Dynamically determine registry path from environment or default."""
    if os.environ.get("SARI_REGISTRY_FILE"):
        return Path(os.environ["SARI_REGISTRY_FILE"]).resolve()
    return REGISTRY_FILE

class ServerRegistry:
    """
    Manages the 'server.json' registry for Sari Daemons.
    Registry v2 (SSOT):
      - daemons: boot_id -> {host, port, pid, start_ts, last_seen_ts, draining, version}
      - workspaces: workspace_root -> {boot_id, last_active_ts, http_port, http_host}
    """

    VERSION = "2.0"

    def __init__(self):
        reg_file = get_registry_path()
        reg_file.parent.mkdir(parents=True, exist_ok=True)
        if not reg_file.exists():
            self._save(self._empty())

    def _empty(self) -> Dict[str, Any]:
        return {"version": self.VERSION, "daemons": {}, "workspaces": {}}

    def _normalize_workspace_root(self, workspace_root: str) -> str:
        from sari.core.workspace import WorkspaceManager
        return WorkspaceManager.normalize_path(workspace_root)

    def _safe_load(self, f) -> Dict[str, Any]:
        try:
            data = json.load(f)
        except Exception:
            data = {}
        return self._ensure_v2(data)

    @contextmanager
    def _registry_lock(self, exclusive: bool):
        reg_file = get_registry_path()
        lock_file = reg_file.with_suffix(".json.lock")
        reg_file.parent.mkdir(parents=True, exist_ok=True)
        with open(lock_file, "a+") as lockf:
            if not IS_WINDOWS:
                fcntl.flock(lockf, fcntl.LOCK_EX if exclusive else fcntl.LOCK_SH)
            try:
                yield
            finally:
                if not IS_WINDOWS:
                    fcntl.flock(lockf, fcntl.LOCK_UN)

    def _load_unlocked(self) -> Dict[str, Any]:
        reg_file = get_registry_path()
        try:
            with open(reg_file, "r") as f:
                return self._safe_load(f)
        except FileNotFoundError:
            return self._empty()

    def _atomic_write(self, data: Dict[str, Any]) -> None:
        reg_file = get_registry_path()
        reg_file.parent.mkdir(parents=True, exist_ok=True)
        tmp_path = reg_file.parent / f"{reg_file.name}.tmp.{os.getpid()}.{int(time.time() * 1000)}"
        try:
            with open(tmp_path, "w") as f:
                json.dump(data, f, indent=2)
                f.flush()
                os.fsync(f.fileno())
            os.replace(tmp_path, reg_file)
        except Exception:
            try:
                if tmp_path.exists():
                    tmp_path.unlink()
            except Exception:
                pass
            raise

    def _ensure_v2(self, data: Dict[str, Any]) -> Dict[str, Any]:
        if not isinstance(data, dict):
            return self._empty()
        version = data.get("version")
        if version == self.VERSION and "daemons" in data and "workspaces" in data:
            return data

        # Migrate legacy schema (v1)
        if "instances" in data and isinstance(data["instances"], dict):
            now = time.time()
            migrated = self._empty()
            for ws, info in data["instances"].items():
                if not isinstance(info, dict): continue
                try:
                    ws_norm = self._normalize_workspace_root(ws)
                except Exception:
                    ws_norm = ws
                pid = info.get("pid")
                port = info.get("port")
                if pid is None or port is None: continue
                
                boot_id = f"legacy-{pid}-{port}"
                migrated["daemons"][boot_id] = {
                    "host": "127.0.0.1",
                    "port": int(port),
                    "pid": int(pid),
                    "start_ts": info.get("start_ts") or now,
                    "last_seen_ts": now,
                    "draining": False,
                    "version": info.get("version") or "legacy",
                }
                migrated["workspaces"][ws_norm] = {
                    "boot_id": boot_id,
                    "last_active_ts": now,
                }
            return migrated

        return self._empty()

    def _load(self) -> Dict[str, Any]:
        """Load registry with shared lock."""
        with self._registry_lock(exclusive=False):
            return self._load_unlocked()

    def _save(self, data: Dict[str, Any]):
        """Save registry with exclusive lock."""
        with self._registry_lock(exclusive=True):
            self._atomic_write(data)

    def _update(self, updater) -> None:
        with self._registry_lock(exclusive=True):
            data = self._load_unlocked()
            updater(data)
            self._atomic_write(data)

    def _is_process_alive(self, pid: Optional[int]) -> bool:
        if not pid:
            return False
        try:
            os.kill(pid, 0)  # Signal 0 checks existence
            return True
        except PermissionError:
            # Sandbox/permission-restricted environments may block signal checks.
            return True
        except OSError:
            return False

    def _prune_dead_locked(self, data: Dict[str, Any]) -> None:
        daemons = data.get("daemons", {})
        workspaces = data.get("workspaces", {})
        dead = [boot_id for boot_id, info in daemons.items()
                if not self._is_process_alive(info.get("pid"))]
        for boot_id in dead:
            daemons.pop(boot_id, None)
        if dead:
            workspaces = {
                ws: info for ws, info in workspaces.items()
                if info.get("boot_id") not in dead
            }
        data["daemons"] = daemons
        data["workspaces"] = workspaces

    def register_daemon(self, boot_id: str, host: str, port: int, pid: int, version: str = "") -> None:
        """Register a running daemon."""
        def _upd(data):
            self._prune_dead_locked(data)
            daemons = data.get("daemons", {})
            daemons[boot_id] = {
                "host": host,
                "port": port,
                "pid": pid,
                "start_ts": daemons.get(boot_id, {}).get("start_ts") or time.time(),
                "last_seen_ts": time.time(),
                "draining": bool(daemons.get(boot_id, {}).get("draining", False)),
                "version": version or daemons.get(boot_id, {}).get("version", ""),
            }
            data["daemons"] = daemons
            data["version"] = self.VERSION
        self._update(_upd)

    def touch_daemon(self, boot_id: str) -> None:
        """Update last_seen for daemon."""
        def _upd(data):
            daemons = data.get("daemons", {})
            if boot_id in daemons:
                daemons[boot_id]["last_seen_ts"] = time.time()
            data["daemons"] = daemons
            data["version"] = self.VERSION
        self._update(_upd)

    def set_daemon_draining(self, boot_id: str, draining: bool = True) -> None:
        def _upd(data):
            daemons = data.get("daemons", {})
            if boot_id in daemons:
                daemons[boot_id]["draining"] = bool(draining)
            data["daemons"] = daemons
            data["version"] = self.VERSION
        self._update(_upd)

    def unregister_daemon(self, boot_id: str) -> None:
        def _upd(data):
            daemons = data.get("daemons", {})
            if boot_id in daemons:
                daemons.pop(boot_id, None)
            workspaces = data.get("workspaces", {})
            workspaces = {
                ws: info for ws, info in workspaces.items()
                if info.get("boot_id") != boot_id
            }
            data["daemons"] = daemons
            data["workspaces"] = workspaces
            data["version"] = self.VERSION
        self._update(_upd)

    def set_workspace(self, workspace_root: str, boot_id: str,
                      http_port: Optional[int] = None, http_host: Optional[str] = None) -> Optional[str]:
        """Bind workspace to daemon (returns previous boot_id if changed)."""
        ws = self._normalize_workspace_root(workspace_root)
        prev_boot = None

        def _upd(data):
            nonlocal prev_boot
            self._prune_dead_locked(data)
            workspaces = data.get("workspaces", {})
            prev_boot = workspaces.get(ws, {}).get("boot_id")
            payload = dict(workspaces.get(ws, {}))
            payload["boot_id"] = boot_id
            payload["last_active_ts"] = time.time()
            if http_port is not None:
                payload["http_port"] = int(http_port)
            if http_host:
                payload["http_host"] = str(http_host)
            workspaces[ws] = payload
            data["workspaces"] = workspaces
            data["version"] = self.VERSION
            # If ownership changed, mark previous daemon as draining.
            if prev_boot and prev_boot != boot_id:
                daemons = data.get("daemons", {})
                if prev_boot in daemons:
                    daemons[prev_boot]["draining"] = True
                data["daemons"] = daemons
        self._update(_upd)
        return prev_boot

    def touch_workspace(self, workspace_root: str) -> None:
        ws = self._normalize_workspace_root(workspace_root)

        def _upd(data):
            workspaces = data.get("workspaces", {})
            if ws in workspaces:
                workspaces[ws]["last_active_ts"] = time.time()
            data["workspaces"] = workspaces
            data["version"] = self.VERSION
        self._update(_upd)

    def set_workspace_http(self, workspace_root: str, http_port: int, http_host: Optional[str] = None, http_pid: Optional[int] = None) -> None:
        ws = self._normalize_workspace_root(workspace_root)

        def _upd(data):
            workspaces = data.get("workspaces", {})
            payload = dict(workspaces.get(ws, {}))
            payload["http_port"] = int(http_port)
            if http_host:
                payload["http_host"] = str(http_host)
            if http_pid:
                payload["http_pid"] = int(http_pid)
            payload["last_active_ts"] = time.time()
            workspaces[ws] = payload
            data["workspaces"] = workspaces
            data["version"] = self.VERSION
        self._update(_upd)

    def unregister_workspace(self, workspace_root: str, boot_id: Optional[str] = None) -> None:
        ws = self._normalize_workspace_root(workspace_root)

        def _upd(data):
            workspaces = data.get("workspaces", {})
            if ws in workspaces:
                if boot_id and workspaces[ws].get("boot_id") != boot_id:
                    return
                workspaces.pop(ws, None)
            data["workspaces"] = workspaces
            data["version"] = self.VERSION
        self._update(_upd)

    def get_daemon(self, boot_id: str) -> Optional[Dict[str, Any]]:
        data = self._load()
        daemon = (data.get("daemons") or {}).get(boot_id)
        if not daemon:
            return None
        if not self._is_process_alive(daemon.get("pid")):
            self.unregister_daemon(boot_id)
            return None
        return daemon

    def get_workspace(self, workspace_root: str) -> Optional[Dict[str, Any]]:
        ws = self._normalize_workspace_root(workspace_root)
        data = self._load()
        return (data.get("workspaces") or {}).get(ws)

    def list_workspaces_for_boot(self, boot_id: str) -> Iterable[str]:
        data = self._load()
        workspaces = data.get("workspaces", {})
        return [ws for ws, info in workspaces.items() if info.get("boot_id") == boot_id]

    def resolve_workspace_daemon(self, workspace_root: str) -> Optional[Dict[str, Any]]:
        ws = self._normalize_workspace_root(workspace_root)
        data = self._load()
        workspaces = data.get("workspaces", {})
        info = workspaces.get(ws)
        if not info:
            return None
        boot_id = info.get("boot_id")
        daemon = (data.get("daemons") or {}).get(boot_id)
        if not daemon:
            self.unregister_workspace(ws)
            return None
        if not self._is_process_alive(daemon.get("pid")):
            self.unregister_daemon(boot_id)
            return None
        merged = dict(daemon)
        merged["boot_id"] = boot_id
        return merged

    def get_instance(self, workspace_root: str) -> Optional[Dict[str, Any]]:
        """Backward compatible alias for resolve_workspace_daemon."""
        return self.resolve_workspace_daemon(workspace_root)

    def register(self, workspace_root: str, port: int, pid: int) -> None:
        """Backward compatible registry (legacy)."""
        boot_id = f"legacy-{pid}-{port}"
        self.register_daemon(boot_id=boot_id, host="127.0.0.1", port=port, pid=pid, version="legacy")
        self.set_workspace(workspace_root, boot_id)

    def unregister(self, workspace_root: str) -> None:
        """Backward compatible unregister."""
        self.unregister_workspace(workspace_root)

    def find_free_port(self, start_port: int = 47777, max_port: int = 65535) -> int:
        """Find a port not in use by other instances AND OS."""
        data = self._load()
        used_ports = {
            info.get("port")
            for info in (data.get("daemons") or {}).values()
            if self._is_process_alive(info.get("pid"))
        }

        for port in range(start_port, max_port + 1):
            if port in used_ports:
                continue
            try:
                with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
                    s.bind(("127.0.0.1", port))
                    return port
            except OSError:
                continue

        raise RuntimeError("No free ports available")



import sys
import json
import socket
import threading
import os
import time
import subprocess
import logging
import sys
import tempfile
import secrets
from pathlib import Path

# Add project root to sys.path for absolute imports
SCRIPT_DIR = Path(__file__).parent
REPO_ROOT = SCRIPT_DIR.parent
if str(REPO_ROOT) not in sys.path:
    sys.path.insert(0, str(REPO_ROOT))

from sari.mcp.telemetry import TelemetryLogger
from sari.core.workspace import WorkspaceManager
from sari.core.server_registry import ServerRegistry
from sari.core.daemon_resolver import resolve_daemon_address as _resolve_daemon_target

try:
    import fcntl  # type: ignore
except Exception:
    fcntl = None

# Configure logging to stderr so it doesn't interfere with MCP STDIO
logging.basicConfig(
    level=logging.INFO,
    format='%(asctime)s - %(name)s - %(levelname)s - %(message)s',
    stream=sys.stderr
)
logger = logging.getLogger("mcp-proxy")
telemetry = TelemetryLogger(WorkspaceManager.get_global_log_dir())


def _log_info(message: str) -> None:
    logger.info(message)
    try:
        telemetry.log_info(message)
    except Exception:
        pass


def _log_error(message: str) -> None:
    logger.error(message)
    try:
        telemetry.log_error(message)
    except Exception:
        pass

DEFAULT_HOST = "127.0.0.1"
DEFAULT_PORT = 47779
MAX_MESSAGE_SIZE = 10 * 1024 * 1024 # 10MB
_HEADER_SEP = b"\r\n\r\n"
_MODE_FRAMED = "framed"
_MODE_JSONL = "jsonl"


def _identify_sari_daemon(host: str, port: int, timeout: float = 0.3) -> bool:
    try:
        with socket.create_connection((host, port), timeout=timeout) as sock:
            sock.settimeout(timeout)
            body = json.dumps({"jsonrpc": "2.0", "id": 1, "method": "sari/identify"}).encode("utf-8")
            header = f"Content-Length: {len(body)}\r\n\r\n".encode("ascii")
            sock.sendall(header + body)

            f = sock.makefile("rb")
            headers = {}
            while True:
                line = f.readline()
                if not line:
                    return False
                line = line.strip()
                if not line:
                    break
                if b":" in line:
                    k, v = line.split(b":", 1)
                    headers[k.strip().lower()] = v.strip()

            try:
                content_length = int(headers.get(b"content-length", b"0"))
            except ValueError:
                return False
            if content_length <= 0:
                return False
            resp_body = f.read(content_length)
            if not resp_body:
                return False
            resp = json.loads(resp_body.decode("utf-8"))
            result = resp.get("result") or {}
            if result.get("name") == "sari":
                return True
    except Exception:
        pass
    return False


def _lock_file(lock_file) -> None:
    if fcntl is not None:
        fcntl.flock(lock_file, fcntl.LOCK_EX)
        return
    try:
        import msvcrt
        lock_file.seek(0)
        msvcrt.locking(lock_file.fileno(), msvcrt.LK_LOCK, 1)
    except Exception:
        pass

def _unlock_file(lock_file) -> None:
    if fcntl is not None:
        fcntl.flock(lock_file, fcntl.LOCK_UN)
        return
    try:
        import msvcrt
        lock_file.seek(0)
        msvcrt.locking(lock_file.fileno(), msvcrt.LK_UNLCK, 1)
    except Exception:
        pass

def start_daemon_if_needed(host, port, workspace_root: str = ""):
    """Checks if daemon is running, if not starts it."""
    if _identify_sari_daemon(host, port):
        return True

    if not workspace_root:
        workspace_root = os.environ.get("SARI_WORKSPACE_ROOT") or WorkspaceManager.resolve_workspace_root()

    lock_path = os.path.join(tempfile.gettempdir(), f"sari-daemon-{host}-{port}.lock")
    with open(lock_path, "w") as lock_file:
        try:
            # Acquire exclusive lock (blocking)
            _lock_file(lock_file)

            # Double-check if daemon started while waiting for lock
            if _identify_sari_daemon(host, port):
                return True

            _log_info("Daemon not running, starting...")

            repo_root = Path(__file__).parent.parent.parent
            env = os.environ.copy()
            env["SARI_DAEMON_AUTOSTART"] = "1"
            env["SARI_WORKSPACE_ROOT"] = workspace_root

            subprocess.Popen(
                [sys.executable, "-m", "sari.mcp.cli", "daemon", "start", "-d"],
                cwd=repo_root,
                env=env,
                start_new_session=True,
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
            )

            # Wait for it to come up (Increase to 10s for cold start reliability)
            for _ in range(100):
                host, port = _resolve_daemon_target()
                if _identify_sari_daemon(host, port):
                    _log_info("Daemon started successfully.")
                    return True
                time.sleep(0.1)

            _log_error("Failed to start daemon.")
            return False

        finally:
            _unlock_file(lock_file)

def forward_socket_to_stdout(sock, state):
    try:
        f = sock.makefile("rb")
        while True:
            # Read Headers
            headers = {}
            while True:
                line = f.readline()
                if not line:
                    break
                line_str = line.decode("utf-8").strip()
                if not line_str:
                    break
                if ":" in line_str:
                    k, v = line_str.split(":", 1)
                    headers[k.strip().lower()] = v.strip()

            if not headers and not line:
                break

            content_length = int(headers.get("content-length", 0))
            if content_length <= 0:
                continue

            body = f.read(content_length)
            if not body:
                break

            msg_id = None
            obj = None
            try:
                obj = json.loads(body.decode("utf-8"))
                if isinstance(obj, dict):
                    msg_id = obj.get("id")
            except Exception:
                msg_id = None
                obj = None

            # If daemon is draining, reconnect and replay initialize without
            # surfacing an error to the client.
            try:
                if isinstance(obj, dict):
                    err = obj.get("error") or {}
                    code = err.get("code")
                    message = str(err.get("message") or "")
                    if code == -32001 and "draining" in message.lower():
                        _log_info("Server draining detected; reconnecting to latest daemon.")
                        state["dead"] = True
                        if _reconnect(state):
                            init_req = state.get("init_request")
                            if init_req:
                                _send_payload(
                                    state,
                                    json.dumps(init_req).encode("utf-8"),
                                    state.get("mode") or _MODE_FRAMED,
                                )
                            continue
            except Exception:
                pass
            if msg_id is not None:
                with state["suppress_lock"]:
                    suppress_ids = state.setdefault("suppress_ids", set())
                    if msg_id in suppress_ids:
                        suppress_ids.discard(msg_id)
                        continue

            mode = state.get("mode") or _MODE_FRAMED
            if mode == _MODE_JSONL:
                sys.stdout.buffer.write(body + b"\n")
                sys.stdout.buffer.flush()
            else:
                header = f"Content-Length: {len(body)}\r\n\r\n".encode("ascii")
                sys.stdout.buffer.write(header + body)
                sys.stdout.buffer.flush()
    except Exception as e:
        _log_error(f"Error forwarding socket to stdout: {e}")
    finally:
        state["dead"] = True
        try:
            sock.close()
        except Exception:
            pass

def _read_mcp_message(stdin):
    """Read one MCP framed message (Content-Length) or JSONL fallback."""
    line = stdin.readline()
    if not line:
        return None
    while line in (b"\n", b"\r\n"):
        line = stdin.readline()
        if not line:
            return None

    # Strict Framing: Only allow JSONL if env var is set
    allow_jsonl = os.environ.get("SARI_DEV_JSONL") == "1"
    if line.lstrip().startswith((b"{", b"[")):
        if not allow_jsonl:
            return None
        return line.rstrip(b"\r\n"), _MODE_JSONL

    headers = [line]
    while True:
        h = stdin.readline()
        if not h:
            return None
        if h in (b"\n", b"\r\n"):
            break
        headers.append(h)

    content_length = None
    for h in headers:
        parts = h.decode("utf-8", errors="ignore").split(":", 1)
        if len(parts) != 2:
            continue
        key = parts[0].strip().lower()
        if key == "content-length":
            try:
                content_length = int(parts[1].strip())
            except ValueError:
                pass
            break

    if content_length is None or content_length <= 0 or content_length > MAX_MESSAGE_SIZE:
        return None

    body = b""
    while len(body) < content_length:
        chunk = stdin.read(content_length - len(body))
        if not chunk:
            break
        body += chunk

    if len(body) < content_length:
        return None
    return body, _MODE_FRAMED


def _send_payload(state, payload: bytes, mode: str) -> None:
    sock = state.get("sock")
    if not sock:
        raise OSError("socket not connected")
    if mode == _MODE_JSONL:
        sock.sendall(payload + b"\n")
    else:
        header = f"Content-Length: {len(payload)}\r\n\r\n".encode("ascii")
        sock.sendall(header + payload)


def _reconnect(state) -> bool:
    max_attempts = int(os.environ.get("SARI_PROXY_RECONNECT_MAX", "10") or 10)
    backoff = float(os.environ.get("SARI_PROXY_RECONNECT_BACKOFF", "0.2") or 0.2)
    with state["conn_lock"]:
        if not state.get("dead") and state.get("sock"):
            return True

        workspace_root = state.get("workspace_root") or ""
        last_err = None
        for _ in range(max_attempts):
            host, port = _resolve_daemon_target()
            if not start_daemon_if_needed(host, port, workspace_root=workspace_root):
                last_err = "daemon start failed"
                time.sleep(backoff)
                continue
            host, port = _resolve_daemon_target()
            try:
                sock = socket.create_connection((host, port))
                last_err = None
                break
            except Exception as e:
                last_err = str(e)
                time.sleep(backoff)
        else:
            _log_error(f"Reconnect failed: {last_err}")
            return False

        state["sock"] = sock
        state["dead"] = False

        t = threading.Thread(target=forward_socket_to_stdout, args=(sock, state), daemon=True)
        t.start()

        init_req = state.get("init_request")
        if init_req:
            try:
                internal = dict(init_req)
                with state["suppress_lock"]:
                    suppress_ids = state.setdefault("suppress_ids", set())
                    internal_id = -secrets.randbelow(2**31 - 1) - 1
                    while internal_id in suppress_ids:
                        internal_id = -secrets.randbelow(2**31 - 1) - 1
                    suppress_ids.add(internal_id)
                internal["id"] = internal_id
                _send_payload(state, json.dumps(internal).encode("utf-8"), state.get("mode") or _MODE_FRAMED)
            except Exception as e:
                _log_error(f"Reconnect initialize failed: {e}")
        return True


def forward_stdin_to_socket(state):
    try:
        stdin = sys.stdin.buffer
        while True:
            res = _read_mcp_message(stdin)
            if res is None:
                break
            msg, mode = res
            # Inject rootUri for initialize when client omits it (per-connection workspace)
            try:
                req = json.loads(msg.decode("utf-8"))

                def _inject(obj):
                    if not isinstance(obj, dict) or obj.get("method") != "initialize":
                        return obj, False
                    params = obj.get("params") or {}
                    if params.get("rootUri") or params.get("rootPath"):
                        state["init_request"] = obj
                        return obj, False

                    # PRIORITY: SARI_
                    ws = None
                    val = os.environ.get("SARI_WORKSPACE_ROOT")
                    if val:
                        ws = val
                    
                    if not ws:
                        state["init_request"] = obj
                        return obj, False
                    params = dict(params)
                    params["rootUri"] = f"file://{ws}"
                    obj = dict(obj)
                    obj["params"] = params
                    _log_info(f"Injected rootUri for initialize: {ws}")
                    state["init_request"] = obj
                    return obj, True

                injected = False
                if isinstance(req, dict):
                    req, injected = _inject(req)
                elif isinstance(req, list):
                    new_list = []
                    for item in req:
                        item2, did = _inject(item)
                        injected = injected or did
                        new_list.append(item2)
                    req = new_list

                if injected:
                    msg = json.dumps(req).encode("utf-8")
            except Exception:
                pass
            if state.get("mode") is None:
                state["mode"] = mode

            if state.get("dead"):
                if not _reconnect(state):
                    return
            try:
                with state["send_lock"]:
                    _send_payload(state, msg, mode)
            except Exception:
                state["dead"] = True
                if not _reconnect(state):
                    return
                with state["send_lock"]:
                    _send_payload(state, msg, mode)
    except Exception as e:
        _log_error(f"Error forwarding stdin to socket: {e}")
        try:
            sock = state.get("sock")
            if sock:
                sock.close()
        except Exception:
            pass

def main():
    # Log startup context for diagnostics
    _log_info(
        "Proxy startup: cwd=%s argv=%s SARI_ROOT=%s"
        % (
            os.getcwd(),
            sys.argv,
            os.environ.get("SARI_WORKSPACE_ROOT"),
        )
    )

    workspace_root = os.environ.get("SARI_WORKSPACE_ROOT") or WorkspaceManager.resolve_workspace_root()
    host, port = _resolve_daemon_target()

    if not start_daemon_if_needed(host, port, workspace_root=workspace_root):
        sys.exit(1)

    try:
        host, port = _resolve_daemon_target()
        sock = socket.create_connection((host, port))
    except Exception as e:
        _log_error(f"Could not connect to daemon: {e}")
        sys.exit(1)

    # Start threads for bidirectional forwarding
    state = {
        "mode": None,
        "sock": sock,
        "dead": False,
        "send_lock": threading.Lock(),
        "conn_lock": threading.Lock(),
        "suppress_lock": threading.Lock(),
        "suppress_ids": set(),
        "init_request": None,
        "workspace_root": workspace_root,
    }
    t1 = threading.Thread(target=forward_socket_to_stdout, args=(sock, state), daemon=True)
    t1.start()

    forward_stdin_to_socket(state)
    t1.join(timeout=1.0)

if __name__ == "__main__":
    main()



import logging
import os
from pathlib import Path
from typing import Optional

from sari.core.settings import settings

def get_logger(name: str, log_file: Optional[str] = None, level: int = logging.INFO) -> logging.Logger:
    logger = logging.getLogger(name)
    logger.setLevel(level)
    
    if not logger.handlers:
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        
        # Stream Handler
        sh = logging.StreamHandler()
        sh.setFormatter(formatter)
        logger.addHandler(sh)
        
        # File Handler if path provided
        if log_file:
            try:
                from logging.handlers import RotatingFileHandler
                log_path = Path(log_file)
                log_path.parent.mkdir(parents=True, exist_ok=True)
                # Max 10MB per file, keep 5 backups
                fh = RotatingFileHandler(log_file, maxBytes=10*1024*1024, backupCount=5)
                fh.setFormatter(formatter)
                logger.addHandler(fh)
            except Exception:
                pass
                
    return logger

def setup_global_logging():
    log_level_str = settings.LOG_LEVEL
    level = getattr(logging, log_level_str, logging.INFO)
    logging.basicConfig(
        level=level,
        format='%(asctime)s - %(name)s - %(levelname)s - %(message)s'
    )



import json
import pytest
import os
import zlib
import time
import urllib.parse
from pathlib import Path
from unittest.mock import MagicMock, patch
from sari.core.indexer.worker import IndexWorker
from sari.core.indexer.main import Indexer
from sari.mcp.server import LocalSearchMCPServer
from sari.core.db import LocalSearchDB

class TestConfigMatrix:
    
    @pytest.fixture
    def mock_db(self, tmp_path):
        db_path = tmp_path / "test.db"
        return LocalSearchDB(str(db_path))

    @pytest.fixture
    def mock_cfg(self):
        cfg = MagicMock()
        cfg.workspace_roots = ["/tmp"]
        cfg.store_content = True
        return cfg

    # 1. Storage Matrix: STORE_CONTENT + COMPRESS
    def test_matrix_storage_compressed(self, mock_db, mock_cfg):
        from sari.core.settings import Settings
        
        # Scenario: Content Stored and Compressed
        with patch.dict("os.environ", {
            "SARI_STORE_CONTENT": "1",
            "SARI_STORE_CONTENT_COMPRESS": "1",
            "SARI_STORE_CONTENT_COMPRESS_LEVEL": "9"
        }):
            s = Settings()
            worker = IndexWorker(mock_cfg, mock_db, None, lambda p, c: ([], []), settings_obj=s)
            
            root = Path("/tmp/test_root_matrix")
            root.mkdir(parents=True, exist_ok=True)
            f = root / "hello.txt"
            content = "Hello " * 100
            f.write_text(content)
            
            res = worker.process_file_task(root, f, f.stat(), int(time.time()), time.time(), False)
            
            assert res["type"] == "changed"
            assert isinstance(res["content"], bytes)
            assert res["content"].startswith(b"ZLIB\0")
            
            # Use real DB connection to verify retrieval
            # First insert the root to satisfy FK
            root_id = res["rel"].split("/")[0]
            mock_db.upsert_root(root_id, str(root), str(root), label="test")
            
            conn = mock_db._write
            cur = conn.cursor()
            mock_db.upsert_files_tx(cur, [(
                res["rel"], "hello.txt", root_id, "repo", res["mtime"], res["size"], 
                res["content"], res["content_hash"], "", 0, 0, "ok", "", "ok", "", 0, 0, 0, res["content_bytes"], res["metadata_json"]
            )])
            conn.commit()
            
            read_back = mock_db.read_file(res["rel"])
            assert read_back == content

    # 2. Search Policy Matrix: SEARCH_FIRST_MODE (warn vs enforce)
    def test_matrix_policy_enforcement(self):
        from sari.mcp.policies import PolicyEngine
        
        # Enforce Mode
        engine = PolicyEngine(mode="enforce")
        # Should block read without search
        res = engine.check_pre_call("read_file")
        assert res is not None
        # Check text in the response (URL encoded in PACK1)
        text = str(res)
        assert "search-first" in text.lower()
        
        # Mark search done
        engine.mark_action("search")
        res = engine.check_pre_call("read_file")
        assert res is None # Allowed now

    # 3. Filter Matrix: MAX_PARSE_BYTES
    def test_matrix_size_filtering(self, mock_db, mock_cfg):
        from sari.core.settings import Settings
        
        with patch.dict("os.environ", {
            "SARI_MAX_PARSE_BYTES": "5"
        }):
            s = Settings()
            worker = IndexWorker(mock_cfg, mock_db, None, lambda p, c: ([], []), settings_obj=s)
            
            root = Path("/tmp/test_root_size_matrix")
            root.mkdir(parents=True, exist_ok=True)
            f = root / "large.txt"
            f.write_text("1234567890")
            
            res = worker.process_file_task(root, f, f.stat(), 0, 0, False)
            assert res["parse_reason"] == "too_large"
            assert res["content"] == ""



import pytest
import json
import time
from unittest.mock import MagicMock
from sari.mcp.tools.search import execute_search
from sari.mcp.tools.list_files import execute_list_files
from sari.mcp.tools.status import execute_status
from sari.core.models import SearchHit, SearchOptions

def test_execute_search_basic():
    db = MagicMock()
    logger = MagicMock()
    engine = MagicMock()
    roots = ["/tmp/ws"]
    hit = SearchHit(repo="repo1", path="path1", score=1.0, snippet="hi")
    engine.search_v2.return_value = ([hit], {"total": 1, "total_mode": "exact"})
    engine.status.return_value = MagicMock(engine_mode="embedded", engine_ready=True, index_version="v1")
    args = {"query": "test", "limit": 10}
    resp = execute_search(args, db, logger, roots, engine=engine)
    assert "PACK1 tool=search ok=true" in resp["content"][0]["text"]
    assert "r:path=path1 repo=repo1" in resp["content"][0]["text"]

def test_execute_list_files_summary():
    db = MagicMock()
    logger = MagicMock()
    roots = ["/tmp/ws"]
    db.get_repo_stats.return_value = {"repo1": 10, "repo2": 5}
    args = {} # summary_only
    resp = execute_list_files(args, db, logger, roots)
    text = resp["content"][0]["text"]
    assert "PACK1 tool=list_files ok=true mode=summary" in text
    assert "r:repo=repo1 file_count=10" in text
    assert "r:repo=repo2 file_count=5" in text

def test_execute_list_files_detail():
    db = MagicMock()
    logger = MagicMock()
    roots = ["/tmp/ws"]
    db.list_files.return_value = ([{"path": "file1", "repo": "repo1", "mtime": 100}], {"total": 1})
    args = {"repo": "repo1", "limit": 10}
    resp = execute_list_files(args, db, logger, roots)
    assert "PACK1 tool=list_files ok=true" in resp["content"][0]["text"]
    assert "p:file1" in resp["content"][0]["text"]

def test_execute_status():
    indexer = MagicMock()
    indexer.status.index_ready = True
    indexer.status.scanned_files = 100
    indexer.status.indexed_files = 80
    indexer.status.errors = 0
    indexer.get_last_commit_ts.return_value = 12345
    
    db = MagicMock()
    db.fts_enabled = True
    db.count_failed_tasks.return_value = (0, 0)
    
    cfg = MagicMock()
    cfg.include_ext = [".py"]
    cfg.http_api_port = 47777
    cfg.workspace_roots = ["/tmp/ws"]
    
    logger = MagicMock()
    
    args = {"details": True}
    resp = execute_status(args, indexer, db, cfg, "/tmp/ws", "0.0.1", logger)
    text = resp["content"][0]["text"]
    assert "PACK1 tool=status ok=true" in text
    assert "m:index_ready=true" in text
    assert "m:scanned_files=100" in text
    assert "m:cfg_include_ext=.py" in text



import threading
import logging
import time
import os
from typing import Dict, Optional, Any
from sari.core.config.manager import ConfigManager
from sari.core.db.main import LocalSearchDB
from sari.core.indexer import Indexer
from sari.core.search_engine import SearchEngine
from sari.core.workspace import WorkspaceManager

logger = logging.getLogger("sari.registry")

class SharedState:
    def __init__(self, workspace_root: str):
        self.workspace_root = WorkspaceManager.normalize_path(workspace_root)
        try:
            from sari.core.settings import settings
            WorkspaceManager.set_settings(settings)
        except Exception:
            pass
        self.root_id = WorkspaceManager.root_id_for_workspace(workspace_root)
        
        # 1. Load Layered Config (Phase 1)
        self.config_manager = ConfigManager(workspace_root)
        self.config_data = self.config_manager.resolve_final_config()
        
        # 2. Init DB (Phase 4)
        db_path = str(WorkspaceManager.get_global_db_path())
        self.db = LocalSearchDB(db_path)
        try:
            from sari.core.settings import settings
            self.db.set_settings(settings)
        except Exception:
            pass
        
        # 3. Init Indexer (Phase 3 & 2)
        from sari.core.config import Config
        defaults = Config.get_defaults(workspace_root)
        defaults["workspace_roots"] = [workspace_root]
        defaults["include_ext"] = self.config_data.get("final_extensions", defaults["include_ext"])
        defaults["include_files"] = self.config_data.get("final_filenames", defaults["include_files"])
        defaults["exclude_dirs"] = self.config_data.get("final_exclude_dirs", defaults["exclude_dirs"])
        defaults["exclude_globs"] = self.config_data.get("final_exclude_globs", defaults["exclude_globs"])
        defaults["gitignore_lines"] = self.config_data.get("gitignore_lines", defaults.get("gitignore_lines", []))
        cfg_obj = Config(**defaults)
        
        self.indexer = Indexer(cfg_obj, self.db, logger=logger)
        
        # 4. Wire Scheduling Coordinator for Read-Priority
        self.db.coordinator = self.indexer.coordinator
        
        try:
            from sari.core.engine_registry import get_default_engine
            self.db.set_engine(get_default_engine(self.db, cfg_obj, cfg_obj.workspace_roots))
        except Exception:
            self.search_engine = SearchEngine(self.db)
            self.db.set_engine(self.search_engine)
        
        # 5. MCP Server instance (Lazy import to avoid circularity)
        from sari.mcp.server import LocalSearchMCPServer
        self.server = LocalSearchMCPServer(workspace_root)
        
        self.last_activity = time.time()
        self.ref_count = 0
        self.persistent = False
        self._lock = threading.Lock()

    def start(self): 
        # 1. Start Indexer
        threading.Thread(target=self.indexer.run_forever, daemon=True).start()
        
        # 2. Start HTTP Server (Phase 4)
        try:
            from sari.core.http_server import serve_forever
            from sari.core.server_registry import ServerRegistry
            
            host = self.config_data.get("server_host", "127.0.0.1")
            port = int(self.config_data.get("server_port", 47777))
            
            # Start HTTP server
            httpd, actual_port = serve_forever(
                host, port, self.db, self.indexer, 
                workspace_root=self.workspace_root,
                mcp_server=self.server
            )
            self.httpd = httpd
            self.http_port = actual_port
            self.http_host = host
            
            # 3. Register HTTP info in ServerRegistry for CLI to find
            ServerRegistry().set_workspace_http(
                self.workspace_root, 
                actual_port, 
                http_host=host, 
                http_pid=os.getpid()
            )
        except Exception as e:
            logger.error(f"Failed to start HTTP server for {self.workspace_root}: {e}")
    
    def touch(self):
        with self._lock:
            self.last_activity = time.time()
        
    def stop(self):
        self.indexer.stop()
        
        # 1. Shutdown HTTP Server
        if hasattr(self, "httpd"):
            try:
                self.httpd.shutdown()
                self.httpd.server_close()
                logger.info(f"HTTP server for {self.workspace_root} stopped.")
            except Exception as e:
                logger.error(f"Error stopping HTTP server: {e}")

        # 2. Cleanup Registry
        try:
            from sari.core.server_registry import ServerRegistry
            ServerRegistry().unregister_workspace(self.workspace_root)
        except Exception:
            pass

        # 3. Close DB
        try:
            self.db.close_all()
        except Exception:
            try: self.db.close()
            except: pass

class Registry:
    _instance = None
    _lock = threading.Lock()
    def __init__(self): self._sessions: Dict[str, SharedState] = {}
    @classmethod
    def get_instance(cls):
        if cls._instance is None:
            with cls._lock:
                if cls._instance is None: cls._instance = cls()
        return cls._instance
    def get_or_create(self, workspace_root: str, persistent: bool = False) -> SharedState:
        with self._lock:
            if workspace_root not in self._sessions:
                session = SharedState(workspace_root)
                session.persistent = bool(persistent)
                session.start()
                self._sessions[workspace_root] = session
            elif persistent:
                self._sessions[workspace_root].persistent = True
            self._sessions[workspace_root].ref_count += 1
            self._sessions[workspace_root].touch()
            return self._sessions[workspace_root]
            
    def touch_workspace(self, workspace_root: str):
        with self._lock:
            if workspace_root in self._sessions:
                self._sessions[workspace_root].touch()
                
    def release(self, workspace_root: str):
        with self._lock:
            if workspace_root in self._sessions:
                state = self._sessions[workspace_root]
                state.ref_count = max(0, state.ref_count - 1)
                state.touch()
                if state.ref_count == 0 and not state.persistent:
                    state.stop()
                    del self._sessions[workspace_root]

    def shutdown_all(self):
        with self._lock:
            for s in self._sessions.values(): s.stop()
            self._sessions.clear()
    def active_count(self) -> int:
        with self._lock:
            return sum(1 for s in self._sessions.values() if s.ref_count > 0)
    def get_last_activity_ts(self) -> float: 
        with self._lock:
            return max((s.last_activity for s in self._sessions.values() if s.ref_count > 0), default=0.0)



from unittest.mock import MagicMock

from sari.mcp.tools.doctor import (
    _check_db_migration_safety,
    _check_storage_switch_guard,
    _check_writer_health,
)


def test_doctor_reports_non_destructive_db_migration_policy():
    res = _check_db_migration_safety()
    assert res["name"] == "DB Migration Safety"
    assert res["passed"] is True


def test_doctor_storage_switch_guard_reports_blocked_state():
    from sari.core.db.storage import GlobalStorageManager

    old_reason = GlobalStorageManager._last_switch_block_reason
    old_ts = GlobalStorageManager._last_switch_block_ts
    try:
        GlobalStorageManager._last_switch_block_reason = "previous writer did not stop cleanly"
        GlobalStorageManager._last_switch_block_ts = 1.0
        res = _check_storage_switch_guard()
        assert res["name"] == "Storage Switch Guard"
        assert res["passed"] is False
        assert "blocked" in res["error"]
    finally:
        GlobalStorageManager._last_switch_block_reason = old_reason
        GlobalStorageManager._last_switch_block_ts = old_ts


def test_doctor_writer_health_without_storage_instance_is_safe():
    from sari.core.db.storage import GlobalStorageManager

    old_instance = GlobalStorageManager._instance
    try:
        GlobalStorageManager._instance = None
        res = _check_writer_health(MagicMock())
        assert res["name"] == "Writer Health"
        assert res["passed"] is True
    finally:
        GlobalStorageManager._instance = old_instance



import os
import json
import socket
import time
import threading
import shutil
import pytest
from pathlib import Path
from unittest.mock import MagicMock, patch
from sari.mcp.daemon import SariDaemon
from sari.mcp.session import Session
from sari.mcp.workspace_registry import Registry

def test_session_lifecycle():
    """Test session initialization and cleanup."""
    temp_dir = Path("/tmp/sari_session_test").resolve()
    if temp_dir.exists(): shutil.rmtree(temp_dir)
    temp_dir.mkdir(parents=True)
    
    # Mocking Registry to avoid DB creation in test
    registry = Registry.get_instance()
    
    # 1. Create Mock Session
    reader = MagicMock()
    writer = MagicMock()
    session = Session(reader, writer)
    
    assert session.running is True
    assert session.workspace_root is None
    
    # 2. Cleanup
    session.cleanup()
    assert session.workspace_root is None
    
    shutil.rmtree(temp_dir)

def test_daemon_server_communication():
    """Test SariDaemon internal server and command handling."""
    port = 49991
    daemon = SariDaemon(host="127.0.0.1", port=port)
    
    # Mocking background indexing
    with patch("sari.core.indexer.main.Indexer.run_forever"):
        daemon_thread = threading.Thread(target=daemon.start, daemon=True)
        daemon_thread.start()
        time.sleep(1.0)
        
        try:
            with socket.create_connection(("127.0.0.1", port), timeout=2.0) as sock:
                req = {"jsonrpc": "2.0", "id": 1, "method": "initialize", "params": {}}
                body = json.dumps(req).encode("utf-8")
                header = f"Content-Length: {len(body)}\r\n\r\n".encode("ascii")
                sock.sendall(header + body)
                
                resp_data = sock.recv(4096)
                assert b'"result":' in resp_data
        finally:
            daemon.stop()
            daemon_thread.join(timeout=2.0)

def test_workspace_registry_singleton():
    """Test that Registry follows singleton pattern and manages state."""
    r1 = Registry.get_instance()
    r2 = Registry.get_instance()
    assert r1 is r2
    
    # Registry uses a shared dict for SharedState
    ws_path = "/tmp/test_ws"
    # SharedState init will fail without real DB, so we mock get_or_create
    with patch.object(r1, "get_or_create") as mock_get:
        mock_get.return_value = MagicMock()
        state = r1.get_or_create(ws_path)
        assert state is not None



import json
import os
import time
import socket
import pytest
import subprocess
import shutil
from pathlib import Path
from unittest.mock import patch, MagicMock

class TestChaosAndResilience:
    
    @pytest.fixture
    def chaos_env(self, tmp_path):
        fake_home = tmp_path / "chaos_home"
        fake_home.mkdir()
        
        env = os.environ.copy()
        env["HOME"] = str(fake_home)
        env["SARI_REGISTRY_FILE"] = str(tmp_path / "chaos_registry.json")
        env["PYTHONPATH"] = os.getcwd() + ":" + env.get("PYTHONPATH", "")
        env["SARI_DAEMON_AUTOSTART"] = "1"
        env["SARI_DAEMON_PORT"] = "48001"
        
        return env

    def test_chaos_stale_pid_recovery(self, chaos_env):
        pid_dir = Path(chaos_env["HOME"]) / ".local" / "share" / "sari"
        pid_dir.mkdir(parents=True, exist_ok=True)
        pid_file = pid_dir / "daemon.pid"
        pid_file.write_text("999999") 
        
        subprocess.run(
            ["python3", "-m", "sari.mcp.cli", "daemon", "start", "-d", "--daemon-port", "48001"],
            env=chaos_env, check=True
        )
        time.sleep(2.0)
        
        new_pid = int(pid_file.read_text().strip())
        assert new_pid != 999999
        
        status = subprocess.run(
            ["python3", "-m", "sari.mcp.cli", "daemon", "status", "--daemon-port", "48001"],
            env=chaos_env, capture_output=True, text=True
        )
        assert "Running" in status.stdout
        
        subprocess.run(["python3", "-m", "sari.mcp.cli", "daemon", "stop", "--daemon-port", "48001"], env=chaos_env)

    def test_chaos_registry_corruption(self, chaos_env):
        reg_file = Path(chaos_env["SARI_REGISTRY_FILE"])
        reg_file.parent.mkdir(parents=True, exist_ok=True)
        reg_file.write_text("THIS IS NOT JSON!!! CORRUPTED!!!")
        
        subprocess.run(
            ["python3", "-m", "sari.mcp.cli", "daemon", "start", "-d", "--daemon-port", "48002"],
            env=chaos_env, check=True
        )
        time.sleep(2.0)
        
        data = json.loads(reg_file.read_text())
        assert data["version"] == "2.0"
        
        subprocess.run(["python3", "-m", "sari.mcp.cli", "daemon", "stop", "--daemon-port", "48002"], env=chaos_env)

    def test_chaos_recursive_symlink(self, tmp_path, chaos_env):
        from sari.core.indexer.scanner import Scanner
        
        ws = tmp_path / "recursive_ws"
        ws.mkdir()
        sub = ws / "sub"
        sub.mkdir()
        loop = sub / "loop"
        try:
            os.symlink(ws, loop)
        except OSError:
            pytest.skip("Symlinks not supported on this platform")
            
        cfg = MagicMock()
        cfg.exclude_dirs = []
        cfg.settings.FOLLOW_SYMLINKS = True 
        cfg.settings.MAX_DEPTH = 10 # Fix MagicMock comparison
        
        scanner = Scanner(cfg)
        entries = list(scanner.iter_file_entries(ws))
        assert len(entries) < 10

    def test_chaos_db_batch_partial_success(self, chaos_env):
        from sari.core.indexer.db_writer import DBWriter, DbTask
        mock_db = MagicMock()
        mock_conn = MagicMock()
        mock_cur = MagicMock()
        mock_db._write = mock_conn
        mock_conn.cursor.return_value = mock_cur
        
        writer = DBWriter(mock_db, max_batch=2)
        tasks = [
            DbTask(kind="upsert_files", rows=[("p1",)]),
            DbTask(kind="upsert_files", rows=[("p2",)])
        ]
        
        # Manually put tasks into the real queue to avoid task_done mismatch
        for t in tasks:
            writer.enqueue(t)
        
        commit_calls = []
        def mock_commit(): commit_calls.append("commit")
        mock_conn.commit.side_effect = mock_commit
        
        rollback_calls = []
        def mock_rollback(): rollback_calls.append("rollback")
        mock_conn.rollback.side_effect = mock_rollback

        with patch.object(writer, "_process_batch") as mock_proc:
            mock_proc.side_effect = [Exception("Locked"), {"files": 1, "files_paths": ["p1"]}, {"files": 1, "files_paths": ["p2"]}]
            
            # Don't mock the queue entirely, just control loop
            writer._stop.set()
            writer._run()
            
            assert len(rollback_calls) == 1
            assert len(commit_calls) == 2
            assert mock_proc.call_count == 3



import time
import heapq
from dataclasses import dataclass, field
from typing import Dict, Any, Optional, List

@dataclass(order=True)
class ScheduledTask:
    priority: float  # Lower value = Higher priority
    timestamp: float = field(compare=False)
    root_id: str = field(compare=False)
    payload: Any = field(compare=False)

class WeightedFairQueue:
    """
    Combines Weighted Fair Queueing (WFQ) with Aging.
    Ensures fairness between roots AND prevents starvation within a root.
    """
    def __init__(self, age_factor: float = 0.05):
        self._queues: Dict[str, List[ScheduledTask]] = {} # Min-heaps per root
        self._weights: Dict[str, float] = {}
        self._active_roots: List[str] = []
        self._current_idx = 0
        self._age_factor = age_factor
        self._lock = __import__("threading").Lock()

    def set_weight(self, root_id: str, weight: float):
        with self._lock:
            self._weights[root_id] = weight

    def put(self, root_id: str, task: Any, base_priority: float = 10.0):
        with self._lock:
            if root_id not in self._queues:
                self._queues[root_id] = []
                self._active_roots.append(root_id)
            
            stask = ScheduledTask(
                priority=base_priority / self._weights.get(root_id, 1.0),
                timestamp=time.time(),
                root_id=root_id,
                payload=task
            )
            heapq.heappush(self._queues[root_id], stask)

    def get(self) -> Optional[tuple]:
        """Weighted Round-Robin + Internal Aging."""
        with self._lock:
            if not self._active_roots: return None

            start_idx = self._current_idx
            while True:
                root_id = self._active_roots[self._current_idx]
                q = self._queues[root_id]
                
                if q:
                    # Apply aging to all elements in this root's queue
                    # (Simplified: we age the head and then pop it)
                    now = time.time()
                    for task in q:
                        wait_time = now - task.timestamp
                        task.priority -= wait_time * self._age_factor
                    
                    heapq.heapify(q)
                    task = heapq.heappop(q)
                    self._current_idx = (self._current_idx + 1) % len(self._active_roots)
                    return root_id, task.payload
                
                self._current_idx = (self._current_idx + 1) % len(self._active_roots)
                if self._current_idx == start_idx: break
            return None

    def qsize(self) -> int:
        return sum(len(q) for q in self._queues.values())


#!/usr/bin/env python3
"""
Sari guidance tool for LLMs.
Returns a short usage guide to encourage search-first behavior.
"""
from typing import Any, Dict
from sari.mcp.tools._util import mcp_response, pack_header, pack_line, pack_encode_text


def execute_sari_guide(args: Dict[str, Any]) -> Dict[str, Any]:
    text = (
        "💡 Sari - Agentic Search Workflow Guide\n\n"
        "이 도구는 대규모 코드베이스에서 당신의 '외부 기억' 역할을 합니다. "
        "토큰을 낭비하며 파일을 하나씩 열어보기 전에, 아래 순서를 따르면 훨씬 정확하고 빠르게 임무를 완수할 수 있습니다.\n\n"
        "1. [정찰] `search` 또는 `repo_candidates`를 사용해 맥락을 파악하세요.\n"
        "2. [식별] `search_symbols`나 `search_api_endpoints`로 정확한 코드 위치를 찾으세요.\n"
        "3. [획득] `read_symbol`을 통해 필요한 구현부만 콕 집어 읽어보세요. (전체 파일을 읽는 것보다 80% 이상의 컨텍스트를 절약합니다)\n"
        "4. [분석] `get_callers` 또는 `get_implementations`로 코드 간의 연결 고리를 파악하세요.\n\n"
        "핵심 원칙: 먼저 '묻고(Search)', 필요한 것만 '취하고(Select)', 그 다음에 '행동(Act)' 하세요. "
        "이것이 당신의 추론 성능을 최상으로 유지하는 방법입니다.\n"
        "주의: 기본 모드는 경고입니다. 필요 시 search-first 모드를 warn/enforce/off로 조정할 수 있습니다."
    )
    def build_pack() -> str:
        lines = [pack_header("sari_guide", {}, returned=1)]
        lines.append(pack_line("t", single_value=pack_encode_text(text)))
        return "\n".join(lines)

    return mcp_response(
        "sari_guide",
        build_pack,
        lambda: {"content": [{"type": "text", "text": text}]},
    )


from __future__ import annotations

import os
from typing import Any, Callable, Dict, List, Optional, Protocol, Tuple

from sari.core.models import SearchHit, SearchOptions
from sari.core.engine_runtime import EmbeddedEngine, SqliteSearchEngineAdapter
from sari.core.settings import settings


class SearchEngineInterface(Protocol):
    def search_v2(self, opts: SearchOptions) -> Tuple[List[SearchHit], Dict[str, Any]]:
        ...

    def repo_candidates(self, q: str, limit: int = 3, root_ids: Optional[List[str]] = None) -> List[Dict[str, Any]]:
        ...


class EngineRegistry:
    def __init__(self) -> None:
        self._factories: Dict[str, Callable[[Any, Any, Any], SearchEngineInterface]] = {}

    def register(self, name: str, factory: Callable[[Any, Any, Any], SearchEngineInterface]) -> None:
        self._factories[name] = factory

    def create(self, name: str, db: Any, cfg: Any = None, roots: Any = None) -> SearchEngineInterface:
        if name not in self._factories:
            raise KeyError(f"engine not registered: {name}")
        return self._factories[name](db, cfg, roots)

    def default(self, db: Any, cfg: Any = None, roots: Any = None) -> SearchEngineInterface:
        name = default_engine_name(cfg)
        return self.create(name, db, cfg, roots)


_REGISTRY = EngineRegistry()
_REGISTRY.register("sqlite", lambda db, _cfg, _roots: SqliteSearchEngineAdapter(db))
_REGISTRY.register("embedded", lambda db, cfg, roots: EmbeddedEngine(db, cfg, roots or []))


def get_registry() -> EngineRegistry:
    return _REGISTRY


def default_engine_name(cfg: Any = None) -> str:
    if cfg is not None:
        mode = (getattr(cfg, "engine_mode", "") or "").strip().lower()
        if mode in {"embedded", "sqlite"}:
            return mode
    return settings.ENGINE_MODE


def get_default_engine(db: Any, cfg: Any = None, roots: Any = None) -> SearchEngineInterface:
    return _REGISTRY.default(db, cfg, roots)



import pytest
from unittest.mock import MagicMock, patch
from sari.core.engine_runtime import EngineRuntime, EngineRouter, EmbeddedEngine, SqliteSearchEngineAdapter
from sari.core.engine.tantivy_engine import TantivyEngine

def test_engine_router():
    engine1 = MagicMock()
    engine2 = MagicMock()
    router = EngineRouter({"root1": engine1, "root2": engine2})
    
    # Upsert
    docs = [{"doc_id": "root1/f1.py"}, {"doc_id": "root2/f2.js"}]
    router.upsert_documents(docs)
    assert engine1.upsert_documents.called
    assert engine2.upsert_documents.called
    
    # Delete
    router.delete_documents(["root1/f1.py"])
    assert engine1.delete_documents.called
    
    # Search
    engine1.search.return_value = [{"score": 0.9}]
    engine2.search.return_value = [{"score": 0.8}]
    results = router.search("query")
    assert len(results) == 2
    assert results[0]["score"] == 0.9

def test_sqlite_adapter():
    db = MagicMock()
    adapter = SqliteSearchEngineAdapter(db)
    status = adapter.status()
    assert status.engine_mode == "sqlite"
    assert status.engine_ready is True

def test_engine_runtime_init():
    with patch('sari.core.workspace.WorkspaceManager.root_id', return_value="root1"):
        runtime = EngineRuntime(["/tmp/ws"])
        assert runtime.root_ids == ["root1"]
        assert runtime.status().engine_ready is False

def test_embedded_engine_status():
    db = MagicMock()
    cfg = MagicMock()
    with patch('sari.core.workspace.WorkspaceManager.root_id', return_value="root1"):
        with patch('sari.core.engine_runtime.EngineRuntime.initialize'):
            engine = EmbeddedEngine(db, cfg, ["/tmp/ws"])
            status = engine.status()
            assert status.engine_mode == "embedded"


@pytest.mark.gate
def test_tantivy_upsert_accepts_id_key():
    class FakeWriter:
        def __init__(self):
            self.deleted = []
            self.added = []
        def delete_documents(self, field, value):
            self.deleted.append((field, value))
        def add_document(self, doc):
            self.added.append(doc)
        def commit(self):
            pass

    class FakeIndex:
        def __init__(self, writer):
            self._writer = writer
        def writer(self, *_args, **_kwargs):
            return self._writer

    fake_writer = FakeWriter()
    engine = TantivyEngine.__new__(TantivyEngine)
    engine._index = FakeIndex(fake_writer)
    engine._writer = None
    engine._writer_lock = __import__("threading").Lock()
    engine.settings = MagicMock()
    engine.settings.ENGINE_INDEX_MEM_MB = 64
    engine.logger = None

    with patch("sari.core.engine.tantivy_engine.tantivy") as mock_tantivy:
        mock_tantivy.Document.side_effect = lambda **kwargs: kwargs
        engine.upsert_documents([{"id": "root1/file.py", "repo": "repo1"}])

    assert ("path", "root1/file.py") in fake_writer.deleted
    assert len(fake_writer.added) == 1
